# LSB Test Execution Framework
# Download Routines Module (Download.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Download;

use strict;
use Time::Local;

use Common;

# Export symbols
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(
	&download_file &spider
);
#----------------------------------------------------------------------

# How the wget tool should be called for downloading files
#
our $wget = "wget -w 3 --retr-symlinks --timeout=120 --progress=dot:mega";
# -N, --timestamping    to preserve the creation time of the file.
# -w, --wait= SECONDS   wait SECONDS between retrievals
# --retr-symlinks       get the linked-to file instead of the symlink itself
# --timeout             sets the network timeout
# --progress=dot:mega   print less dots
# TODO: ^^ check if this option is available

# How the curl tool should be called.
#
our $curl = "curl -R --retry 3 --retry-delay 3 --connect-timeout 120 --create-dirs -L";
# -R --remote-time  Keep timestamp
# --retry           Specify retry times
# --retry-delay     Specify wait interval between 2 retrys
# --connect-timeout Timeout for a connection
# --create-dirs     Create the necessary local directory hierarchy as needed
# -p --proxytunnel  Ftp protocol tunnel through the http proxy
# -x --proxy        Http proxy
# -U --proxy-user   Specify user and password to use for proxy authentication
# -L --location     Follow redirect
# -#                Linear progress meter

# TODO: add -p on request only, it doesn't work with SQUID!

our $tool = "curl"; # "wget" is an alternative.

our $enabled = 0;

#----------------------------------------------------------------------

# Download $link to file $dest (file name with path).
#
# %options:
#  $options{RESUME} - continue downloading partially-downloaded files.
#  $options{OVERWRITE} - download and overwrite the existing file, even if it looks similar.
#  $options{SPIDER}->{} - Use this data instead of quering information about the file from server.
#
# This function will create the destination directory if it doesn't exist.
#
# Returns an error if failed to download the file.
sub download_file {
	my ($link, $dest, %options) = @_;
	
	( $enabled ) or return error "Downloading isn't enabled.";
	
	my $dest_dir = extract_dir( $dest )
		or return error "Wrong directory of file '$dest'";
	
	# Create the directory if it doesn't exist.
	if ( !-d $dest_dir ) {
		cmd( "mkdir -p ".shq($dest_dir) );
	}
	( -d $dest_dir ) # check again
		or return error "Failed to create directory '$dest_dir'";
	
	my $sp;
	if ( $options{SPIDER} ) {
		$sp = $options{SPIDER};
	} else {
		$sp = spider($link);
	}
	# Do a spider query to receive information about the file from the server
	if ( $sp->{NOTFOUND} ) {
		return error "Can't download '$link':\n The file was removed or replaced.";
	}
	
	# Check if the file is already downloaded
	if ( -f $dest ) {
		my ($size, $mtime) = file_info($dest);
		if ( defined $sp->{SIZE} && defined $size && $sp->{SIZE} == $size
			 && defined $sp->{TIMESTAMP} && $sp->{TIMESTAMP} == $mtime )
		{
			# The file seems to be the same as the one on server
			debug_inform "File '$link' is already downloaded.";
			return 1 unless $options{OVERWRITE};
		}
	}
	
	# The $link will be downloaded to a temporary file,
	# then will be moved to $dest if everything is OK.
	my $temp_file = $dest.".#part";
	
	# Check if the file was partially downloaded
	my $continue = 0;
	if ( -f $temp_file ) {
		my ($tmp_size, $tmp_mtime) = file_info($temp_file);
		# Ensure the file part isn't out-of-date
		if ( $sp->{TIMESTAMP} && $tmp_mtime > $sp->{TIMESTAMP} ) {
			$continue = 1 if $options{RESUME};
		}
		if ( !$continue ) {
			# remove the temporary file if don't want to resume downloading
			cmd("rm -f ".shq($temp_file) );
		}
	}
	
	inform "Downloading '$link'".( $sp->{SIZE} ? " ".$sp->{SIZE}." bytes" : "");
	
	if ( $tool eq 'curl' ) {
		# CURL
		cmd( "cd ".shq($dest_dir)	# just to be safe
			." && $curl ".shq($link)	# download the $link
					.($continue ? " -C -" : "")	# continue downloading if enabled
					." -o ".shq($temp_file)		# save to $temp_file
			) == 0
			or return error "Sorry, failed to download '$link'.", error $last_cmd_output;

	}
	elsif ( $tool eq 'wget' ) {
		# WGET
		cmd( "cd ".shq($dest_dir)	# just to be safe
			." && $wget ".shq($link)	# download the $link
					.($continue ? " -c" : "")	# continue downloading if enabled
					." -O ".shq($temp_file)		# save to $temp_file
			) == 0
			or return error "Sorry, failed to download '$link'.", error $last_cmd_output;
	}
	else { return error "Unexpected download tool: '$tool'"; }
	
	# set the file time
	if ( $sp->{TIMESTAMP} ) { # Set the file modification time as on the server
		# Wget doesn't keep the modification time when is used with the -O key.
		utime $sp->{TIMESTAMP}, $sp->{TIMESTAMP}, $temp_file;
	}
	
	# Check the file size
	if ( $sp->{SIZE} ) {
		if ( -s $temp_file != $sp->{SIZE} ) {
			return error "Size of the downloaded file differs from the expected one.";
		}
	}
	
	# Move the file
	my $result = cmd( "mv -f ".shq($temp_file)." ".shq($dest) );
	if ( $result != 0 || !-f $dest ) { # Failed
		return error "Failed to move the downloaded file '$temp_file' to '$dest'.";
	}
	
	inform "Downloaded file '".extract_filename($dest)."' ".(-s $dest)." bytes";
	
	return 1; # Successfully downloaded
}
#----------------------------------------------------------------------

# Converts time string of format "Wed, 25 Mar 2009 09:26:18 GMT" into unix time numeric value.
# This is used for parsing the HTTP header's "Last-Modified" field.
sub str2time {
	my ($s) = @_;
	
	#     "Wed,    25      Mar     2009     09:26:18          GMT"
	$s =~ /\w+,\s+(\d+)\s+(\w+)\s+(\d+)\s+(\d+):(\d+):(\d+)\s+GMT/ or return undef;
	#             $mday   $MON    $year   $hour $min  $sec
	my ($mday, $MON, $year, $hour, $min, $sec) = ($1, $2, $3, $4, $5, $6);
	
	( $mday >= 1 && $mday <= 31 ) or return undef;
	( $year >= 1900 ) or return undef;
	
	# Find the month number
	my $mon = in_array( lc $MON, qw{jan feb mar apr may jun jul aug sep oct nov dec} ); # from 1 to 12
	( $mon > 0 ) or return undef;
	
	return timegm($sec,$min,$hour,$mday,$mon-1,$year-1900);
}

sub spider {
	if ( $tool eq 'curl' ) {
		return curl_spider(@_);
	}
	elsif ( $tool eq 'wget' ) {
		return wget_spider(@_);
	}
}

sub curl_spider {
	my ($link) = @_;
	
	( $enabled ) or return error "Downloading isn't enabled.";
	
	my $resp = capture("$curl --head ".shq($link));
	
	my $res = {};
	
	my @lines = split /[\r\n]+/, $resp;
	while ( @lines ) {
		my $line = shift @lines;
		
		# curl response looks similar for both http and ftp (many thanks to curl developers)
		if ( $line =~ /^\s*Last-Modified:\s*(.*)/ ) {
			my $dt = $1;
			$res->{TIMESTAMP} = str2time($dt);
		}
		elsif ( $line =~ /^\s*Content-Length:\s*(\d+)/ ) {
			$res->{SIZE} = int($1);
		}
		elsif ( $line =~ /^\s*HTTP\S*\s+(\d+)\b/ ) {
			my $code = $1;
			if ( $code =~ /^[45]\d\d$/ ) { # 4xx - client error, 5xx - server error
				$res->{NOTFOUND} = 1;
			}
		}
	}
	
	return $res;
}

# Get some information about file at a remote server specified by $link.
# Requests a HTTP header using the '--spider' option of wget.
# Returns $res->{SIZE => $size, TIMESTAMP => $last_modified, NOTFOUND => $if_404 }
sub wget_spider {
	my ($link) = @_;
	
	( $enabled ) or return error "Downloading isn't enabled.";
	
	my $resp = capture("$wget -S --spider ".shq($link));
	
	my $res = {};
	
	my $resp_mode = '';
	if ( $link =~ /^https?:/ ) {
		$resp_mode = 'http';
	}
	elsif ( $link =~ /^s?ftp:/ ) {
		$resp_mode = 'ftp';
	}
	else {
		warning "Unsupported protocol: '$link'";
		return $res;
	}
	
	my @lines = split /[\r\n]+/, $resp;
	while ( @lines ) {
		my $line = shift @lines;
		
		if ( $resp_mode eq 'http' ) {
			# Parse an HTTP header line
			if ( $line =~ /^\s+Last-Modified:\s*(.*)/ ) {
				my $dt = $1;
				$res->{TIMESTAMP} = str2time($dt);
			}
			elsif ( $line =~ /^\s+Content-Length:\s*(\d+)/ ) {
				$res->{SIZE} = int($1);
			}
			elsif ( $line =~ /^\s+HTTP\S*\s+(\d+)\b/ ) {
				my $code = $1;
				if ( $code =~ /^[45]\d\d$/ ) { # 4xx - client error, 5xx - server error
					$res->{NOTFOUND} = 1;
				}
			}
		}
		elsif ( $resp_mode eq 'ftp' ) {
			# Parse an FTP response line
			if ( $line =~ /^\s+HTTP\b/ ) {
				# HTTP response detected (probably HTTP-FTP proxy is in use)
				# Switch to parsing HTTP response
				unshift @lines, $line;
				$resp_mode = 'http';
				redo;
			}
			elsif ( $line =~ /^--> SIZE / ) { # file size requested
				$line = $lines[0]; # the next line
				$line = $lines[1] if !$line; # skip a line if it is empty
				if ( $line && $line =~ /^213s+(\d+)/ ) { # 213 - file status code
					$res->{SIZE} = int($1);
				}
			}
		}
	}
	return $res;
}

#----------------------------------------------------------------------
1; # returned value

