# LSB Test Execution Framework
# Package Management Module (Packages.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Packages;
use strict;

use Common;

# Export symbols
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(
	&guess_package_manager
	&package_info &packages_info
	&install_package &remove_package
	&check_file &check_files
);
#----------------------------------------------------------------------

our $package_manager = undef;

#----------------------------------------------------------------------

# Determines which package manager to use.
# Returns "rpm", "dpkg" or undef.
sub guess_package_manager {
	my ($host_os) = @_;
	
	if ( $host_os =~ /(suse|red\W?hat|asianux|mandriva|turbolinux|gentoo)/i ) {
		# Native package manager for Suse and Redhat is rpm
		return "rpm";
	}
	elsif ( $host_os =~ /(ubuntu|debian|freespire|linspire|rays.?lx)/i ) {
		# For Debian and Ubuntu use alien and dpkg.
		return "dpkg";
	}
	else {
		if ( does_shell_know ("dpkg") ) {
			return "dpkg";
		} else {
			return "rpm";
		}
	}
	return undef;
}

#---------------------------------------------------------------------
# Check if rpm package is installed
# Returns version of installed package, or "" if not installed (or errors).
# Used by function package_info().
sub rpm_package_info {
	my ($package_name) = @_;
	
	my $result = "";
	$result = capture( "rpm -q ".shq($package_name) );
	chomp $result;
	if ( $? != 0 ) { $result = ""; }
	
	$result =~ s/^\Q$package_name-\E//; # extract the version number
	
	if ( $result =~ m/\.([^\.]+)$/ ) {  # detach the architecture name
		if ( $1 && guess_architecture($1, $1) ) {
			$result =~ s/\.[^\.]+$//;
		}
	}
	
	return $result;
}

#---------------------------------------------------------------------

# Check if deb package is installed
# Returns version of installed package, or "" if not installed (or errors).
# Used by function package_info().
sub dpkg_package_info {
	my ($package_name) = @_;

	$package_name =~ s/_/-/g; # alien replaces "_" to "-" in package names.

	my $result = "";
	$result = capture( "dpkg -s ".shq($package_name)." 2>/dev/null | grep -e '^\\(Version\\|Status\\):'" );
	chomp $result;
	
	if ( $? == 0 ) {
		if ($result !~ m/Status:\s+\w+\s+ok\s+installed/) { $? = 1; }
	}
	
	if ( $? == 0 ) {
		if ($result =~ m/Version:\s+([^\n]+)/) {
			$result = $1;
		} else {
			warning "Can't understand the answer of dpkg: $result";
			$result = "";
		}
	}
	if ( $? != 0 ) { $result = ""; }
	
	return $result;
}

#---------------------------------------------------------------------

# Returns info about installed package,
# or empty string if the package isn't installed (or in case of errors).
sub package_info {
	my ($package_name) = @_;

	( defined $package_manager ) or fail "Package manager is undefined";
	
	my $result;

	if ( $package_manager eq "rpm" ) {
		$result = rpm_package_info ($package_name);
	}
	elsif ( $package_manager eq "dpkg" ) {
		$result = dpkg_package_info ($package_name);
	}
	
	return $result;
}
#-----------------------------------------------------------------------

# Enquire rpm package manager about state of several packages.
# Returns structure $packages->{$package_name}{VERSIONS}{$package_ver} = 1;
# Used by function packages_info().
sub rpm_packages_info {
	my ($packages) = @_;
	
	my @package_names = ();
	foreach my $name ( sort keys %$packages ) {
		# TODO: skip files, not packages (needs additional field in the $manifest->{PACKAGES} structure)
		push @package_names, $name;
	}
	
	my $resp = capture( "rpm -q ".(join " ", map shq($_), @package_names)." 2>/dev/null" ); 
	my @lines = split(/\n/, $resp);
	
	# rpm -q print packages with versions in the order the packages were requested.
	# There may be several versions of a package installed.
	
	my $result;
	
	while ( @lines ) {
		my $line = shift @lines;
		
		# Package version starts with '-' and a number
		if ( $line =~ s/-(\d\S*)$// ) { # package-name-1.2.3
			my $package_ver = $1;
			my $package_name = $line;
			
			if ( $package_ver =~ m/\.([^\.]+)$/ ) {  # detach the architecture name
				if ( $1 && guess_architecture($1, $1) ) {
					$package_ver =~ s/\.[^\.]+$//;
				}
			}
			
			if ( $packages->{$package_name} ) {
				$packages->{$package_name}{VERSIONS}{$package_ver} = 1;
			}
		}
	}
	
	return $packages;
}

# Enquire dpkg package manager about state of several packages.
# Returns structure $packages->{$package_name}{VERSIONS}{$package_ver} = 1;
# Used by function packages_info().
sub dpkg_packages_info {
	my ($packages) = @_;
	
	# Set COLUMNS=120 to prevent package name chomping.
	my $resp = capture( "COLUMNS=120 dpkg -l ".(join " ", map { shq($_) } sort keys %$packages)." 2>/dev/null" );
	my @lines = split(/\n/, $resp);
	
	# dpkg -l prints a table with installed packages
	
	my $result;
	
	while ( @lines ) {
		my $line = shift @lines;
		
		# Output expected:
		#--------------------------------------------------------------------------------
		# Desired=Unknown/Install/Remove/Purge/Hold
		# | Status=Not/Inst/Cfg-files/Unpacked/Failed-cfg/Half-inst/trig-aWait/Trig-pend
		# |/ Err?=(none)/Hold/Reinst-required/X=both-problems (Status,Err: uppercase=bad)
		# ||/ Name               Version        Description
		# +++-==================-==============-=========================================
		# ii  package_name       1.3.20         ...
		#--------------------------------------------------------------------------------
		# The first letter isn't important (just a desired status)
		# The second letter should be "i" - installed.
		# The third char should be space - no errrors
		if ( $line =~ /^\wi\s\s+(\S+)\s+(\S+)/ ) { 
			my $package_name = $1;
			my $package_ver = $2;
			
			$package_ver =~ s/\.noarch$//; # remove .noarch
			
			if ( $packages->{$package_name} ) {
				$packages->{$package_name}{VERSIONS}{$package_ver} = 1;
			}
		}
	}
	
	return $packages;
}

# Enquire package manager about state of several packages.
# Returns structure $packages->{$package_name}{VERSIONS}{$package_ver} = 1;
sub packages_info {
	my (@params) = @_;
	
	( defined $package_manager ) or return error "Package manager is undefined";
	
	if ( $package_manager eq "rpm" ) {
		return rpm_packages_info (@params);
	}
	elsif ( $package_manager eq "dpkg" ) {
		return dpkg_packages_info (@params);
	}
	
	return undef;
}

#-----------------------------------------------------------------------

# Check if $file matches the expected parameters in the $file_obj structure.
# That are: size (SIZE) and modification time (TIMESTAMP).
# Returns an error if the file doesn't match.
sub check_file {
	my ($file, $file_obj) = @_;
	
	my ($size,$mtime) = file_info($file);
	( defined $size ) or return error "Failed to stat file '$file'";
	
	if ( defined $file_obj->{SIZE} && $file_obj->{SIZE} != 0 ) { # Treat 0 as unknown file size
	# Check the file size
		if ( $size != $file_obj->{SIZE} ) {
			my $filename = extract_filename($file);
			return error "$filename: file size doesn't match: $size != ".$file_obj->{SIZE};
		}
	}
	
	if ( defined $file_obj->{TIMESTAMP} ) {
	# Check the timestamp
		if ( $mtime != $file_obj->{TIMESTAMP} ) {
			my $filename = extract_filename($file);
			return error "$filename: file timestamp doesn't match: $mtime != ".$file_obj->{TIMESTAMP};
		}
	}
	
	return 1; # OK
}
#-----------------------------------------------------------------------

# Check files in cache.
# Set the $file->{DOWNLOADED} field for each file in array @files:
#  -1: the file exists but is invalid (see check_file for criteries).
#   1: the file exists and is valid.
#   0: the file is absent.
sub check_files {
	my (@files) = @_;
	
	foreach my $file ( @files ) {
		my $file_in_cache = cache_dest($file->{SRC})."/".$file->{FILENAME};
		if ( -f $file_in_cache ) {
			if ( is_ok check_file( $file_in_cache, $file )  ) {
				$file->{DOWNLOADED} = 1;
			} else {
				$file->{DOWNLOADED} = -1;
			}
		}
		else {
			$file->{DOWNLOADED} = 0;
		}
	}
	
	return 1;
}

#-----------------------------------------------------------------------

# Install the specified package file.
# The function caller should check the result by calling package_info() then.
sub install_package {
	my ($package_file, $temp_dir) = @_;
	
	# Check that package_file exists
	unless( -f $package_file ) {
		return error "File doesn't exist (or not a file): $package_file";
	}
	
	# Install it.
	inform "Installing '$package_file'";
	
	if( $package_manager eq "rpm" ) {
		
		# Check the file
		if ( $package_file =~ /\.rpm$/ ) {
			cmd( "rpm -i ".shq($package_file) ) == 0
				or inform "!!! Warning: Installation of '$package_file' may be failed";
			return 1; # Caller should check the result by calling package_info().
		}
		# else
		return error "Not an RPM package: '$package_file'";
	}
	elsif( $package_manager eq "dpkg" ) {
		
		# Check the file
		if ( $package_file =~ /\.deb$/ ) {
			cmd( "dpkg -i ".shq($package_file) ) == 0
				or inform "!!! Warning: Installation of '$package_file' may be failed";
			return 1; # Caller should check the result by calling package_info().
		}
		elsif ( $package_file =~ /\.rpm$/ ) {
			# Using alien to convert .rpm to .deb
			
			# An rpm package is first converted to a deb, then installed.
			# It's not Ok to install it straightway using 'alien -ick' command,
			# because some old versions of alien have an issue
			# related to execution of control scripts when -i option is used.
			
			( defined $temp_dir ) or return error "Temporary dir is undefined";
			my $convert_dir = $temp_dir.'/alien-convert';
			is_ok create_empty_directory($convert_dir)
				or return error "Failed to create directory for convertation: '$convert_dir'", $Error::Last;
			my $resp = capture ("cd ".shq($convert_dir)
								." && alien -d -c -k ".shq($package_file) ); 
			if ( $? != 0 ) {
				inform "!!! Warning: Alien may be failed on '$package_file'";
			}
			my $resp_orig = $resp;
			
			# expected response: "package_name.deb generated"
			chomp $resp;
			$resp =~ s/([^\n\x0D]*[\n\x0D])*//sg; # Leave only the last line;
			$resp =~ s/ .*//; # Remove error message if any
			if ( $resp eq "" ) {
				return error "Empty response from alien: '$resp_orig'";
			}
			
			cmd( "dpkg -i ".shq($convert_dir."/".$resp) ) == 0
				or inform "!!! Warning: Installation of '$package_file' may be failed";
			
			return 1; # Caller should check the result by calling package_info().
		}
	}
}
#----------------------------------------------------------------------------

# Uninstall the specified package if it is installed.
# The function caller should check the result by calling package_info() then.
sub remove_package {
	my ($package_name) = @_;

	# Check whether the package is installed.
	if (!package_info($package_name)) {
		# The package seems not to be installed.
		return 1; # OK
	}
	
	# The package seems to be installed, trying to uninstall it.

	# Failure to uninstall package is (mostly) not fatal.
	if( $package_manager eq "rpm" ) {
		cmd( "rpm -e --allmatches --nodeps ".shq($package_name)." 2>&1" );
	}
	elsif( $package_manager eq "dpkg" ) {
		$package_name =~ s/_/-/g; # alien replaces "_" with "-" in package names
		cmd( "dpkg -P ".shq($package_name)." 2>&1" );
	}
	
	return 1; # OK
}
#----------------------------------------------------------------------

#-----------------------------------------------------------------------

return 1;
