#!/usr/bin/perl -w

# LSB Test Execution Framework
# Journal Comparer (jdiff.pl)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use strict;
use FindBin;

sub BEGIN {
	# Add current directory to the @INC to be able to include modules
	# from the same directory where this script is located.
	unshift @INC, $FindBin::Bin;
	
	$| = 1; # Immediate flush of all output
}

use Common;
use Report;

#----------------------------------------------------------------------

my $skip_nondiffers = 1;

my $show_equal = 0;
my $cmp_undef = 1;
my $cmp_content = 1;

my $report_diff = undef;

#----------------------------------------------------------------------

my $box_id = 0;

#----------------------------------------------------------------------

sub my_complain {
	my ($message) = @_;
	
	print STDERR $message."\n";
}

$Common::complain_sub = \&my_complain;

sub guess_journal_type {
	my ($filename) = @_;
	
	if ( $filename =~ /\.journal$/ ) {
		return "TET";
	}
	elsif ( $filename =~ /\.pythonjournal$/ ) {
		return "PYTHON";
	}
	elsif ( $filename =~ /\.tcljournal$/ ) {
		return "TCL";
	}
	elsif ( $filename =~ /\.simplejournal$/ ) {
		return "SIMPLE";
	}
	
	die "Can't understand journal type for '$filename'";
}

sub print_help {
	
	print <<HEREDOC;
Usage: $0 [-r] [-u] [-e] [-c] file1.journal file2.journal file3.journal ...

HEREDOC
	
	print "    -r       Compare reports. Result directories should be specified for this option.\n";
	
	unless ( $show_equal ) {
		print "    -e       Show matching test cases.\n";
	} else {
		print "    -e       Don't show matching test cases.\n";
	}
	unless ( $cmp_undef ) {
		print "    -u       Show 'skipped' test cases.\n";
	} else {
		print "    -u       Don't show 'skipped' test cases.\n";
	}
	unless ( $cmp_content ) {
		print "    -c       Compare test messages.\n";
	} else {
		print "    -c       Don't compare test messages.\n";
	}
}

my @journals = ();
my @journal_names = ();

my @args = @ARGV;

while ( @args ) {
	my $arg = shift @args;
	
	if ( $arg eq "-u" ) {
		$cmp_undef = !$cmp_undef;
	}
	elsif ( $arg eq "-e" ) {
		$show_equal = !$show_equal;
	}
	elsif ( $arg eq "-c" ) {
		$cmp_content = !$cmp_content;
	}
	elsif ( $arg eq "-r" ) {
		$report_diff = 1;
		
		$show_equal = 1;
		$cmp_undef = 1;
		$cmp_content = 1;
	}
	elsif ( $arg eq "--help" ) {
		print_help(); exit;
	}
	elsif ( $arg =~ /^-/ ) {
		print_help(); exit;
	}
	else {
		push @journals, glob($arg);
	}
}

my $hash = {};
my $order = [];
my $number;

is_ok Report::load_problem_database(PROBLEM_DB_FILE)
	or warning "Failed to load problem database", $Error::Last;

if ( @journals == 1 ) {
	# DBG: Start
	#my $jstruct = Report::ParseJournal( $journals[0], guess_journal_type($journals[0]) );
	#is_ok($jstruct) or fail "Failed to parse journal", $Error::Last;
	#
	#use Print_ref;
	#
	#print "Journal structure:\n";
	#print_ref $jstruct;
	# DBG: End
	print_help();
	exit 0;
}
elsif ( @journals == 0 ) {
	print_help(); exit;
}
else {
	$number = @journals;
	
	foreach ( my $n = 0; $n < $number; $n++ ) {
		my $jstruct;
		if ( $report_diff ) {
			$journal_names[$n] = 'Report #'.($n+1);
			$jstruct = make_report_pseudo_jstruct( $journals[$n] )
				or warning("Failed to process ".$journals[$n]), next;
		} else {
			$journal_names[$n] = 'Journal #'.($n+1);
			$jstruct = Report::ParseJournal( $journals[$n], guess_journal_type($journals[$n]) );
			if ( !is_ok($jstruct) ) {
				warning "Failed to parse journal", $Error::Last;
				next;
			}
		}
		
		add_jstruct($jstruct, $n);
	}
	
	print_result();
}

sub make_report_pseudo_jstruct {
	my ($dir) = @_;
	
	my $jstruct;
	
	$dir .= '/results';
	
	my @files = glob( $dir."/*.info" );
	
	foreach my $info_filename ( @files ) {
		# Read the .info file
		my $test_info = read_testinfo( $info_filename );
		if ( !is_ok($test_info) ) {
			warning "Can't read '$info_filename'", $Error::Last;
			next;
		}
		
		foreach my $journal_info ( @{$test_info->{JOURNALS}} ) {
			my $journal_file = $journal_info->{FILE};
			
			my $journal_jstruct = {};
			
			is_ok Report::ParseJournal( $dir."/".$journal_file, $journal_info->{KIND}, $journal_jstruct )
				or warning("Failed to parse journal ".$dir."/".$journal_file), next;
			
			my $stats = $journal_jstruct->{STATS}
				or warning("No stats in '$journal_file'"), next;
			
			my $verdict = Report::verdict( $stats );
			
			my $description = 'Tests Total: '.$stats->{TESTS_TOTAL}."\n"
				.'Passed: '.$stats->{PASSED}."\n"
				.'Failed: '.$stats->{FAILURES}."\n"
				.'Warnings: '.$stats->{WARNINGS}."\n"
				.'FIPs: '.$stats->{FIPS}."\n"
				.'Test suite errors: '.$stats->{TESTSUITE_ERRORS}."\n"
				;
			
			#my $test_name = $test_info->{NAME};
			#if ( @{$test_info->{JOURNALS}} > 1 ) {
				#$test_name .= ' ('.$journal_file.')';
			#}
			my $test_name = $journal_file;
			
			my $jentry = {
					'TEST_NAME' => $test_name,
					'SEVERITY' => $verdict,
					'COMMENT' => $description,
				};
			push @{$jstruct->{ARR}}, $jentry;
		}
	}
	
	return $jstruct;
}

sub add_jstruct {
	my ($jstruct, $n) = @_;

	my $old_order = $order;
	$order = [];
	my $order_size = @$old_order;
		
	my $last_i = 0;
	my $j = 0;
	foreach my $obj ( @{$jstruct->{ARR}} ) {
		my $testcase;
		my $val;
		if ( ref($obj) eq "" ) {
			if ( $obj =~ /^([^:]*):(.*)$/ ) {
				$val = $1;
				$testcase = $2;
			}
		} else {
			( ref($obj) eq "HASH" ) or return error "Wrong entry type in jstruct";
			
			$testcase = $obj->{TEST_NAME};
			$testcase = "***" if !defined $testcase; # testsuite error
			
			$testcase .= " ".$obj->{TEST_POINT} if defined $obj->{TEST_POINT} && $obj->{TEST_POINT} ne "";
			
			$val = $obj;
		}
		
		if ( defined $hash->{$testcase} ) { # known element
			if ( defined $hash->{$testcase}->[$n] ) { # already counted
				if ( $hash->{$testcase}->[$n] eq "?" ) { # unpaired
					$hash->{$testcase}->[$n] = $val;
				} else { # paired
					warning "Double testcase '$testcase'";
				}
			}
			else { # not counted
				# find it in the old order
				for ( my $i = $last_i; $i < $order_size; $i++ ) {
					$last_i = $i+1;
					my $old_testcase = $old_order->[$i];
					if ( $old_testcase ne $testcase ) {
						# add unpaired element
						push @$order, $old_testcase;
						$hash->{$old_testcase}->[$n] = "?"; # Mark as unpaired
					}
					else {
						# stop the search
						last;
					}
				}
				# add paired element
				push @$order, $testcase;
				$hash->{$testcase}->[$n] = $val;
			}
		}
		else { # new element
			push @$order, $testcase;
			$hash->{$testcase}->[$n] = $val;
		}
	}
	continue { $j++ }
	
	for ( my $i = $last_i; $i < $order_size; $i++ ) {
		my $old_testcase = $old_order->[$i];
		if ( !defined $hash->{$old_testcase} 
			|| !defined $hash->{$old_testcase}->[$n] ) 
		{
			# add unpaired element
			push @$order, $old_testcase;
			$hash->{$old_testcase}->[$n] = "?"; # Mark as unpaired
		}
	}
}
#-----------------------------------------------------------------------

sub cmp_jentries ($$) {
	my ( $a, $b ) = @_;
	
	if ( defined $a ) {
		if ( !defined $b ) { return 1 } # not equal
	}
	else {
		if ( defined $b ) { return 1 } # not equal
		else { return 0 } # equal
	}
	
	if ( ref($a) eq "" ) {
		if ( ref($b) ne "" ) { return 1 } # not equal
		
		if ( $a ne $b ) { return 1 }
	}
	else {
		if ( ref($b) eq "" ) { return 1 } # not equal
		
		if ( $cmp_content ) {
			foreach my $key ( 'SEVERITY', 'MESSAGE', 'COMMENT' ) {
				if ( defined $a->{$key} ) {
					if ( !defined $b->{$key} ) { return 1 } # not equal
					if ( $a->{$key} ne $b->{$key} ) { return 1 }
				} else {
					if ( defined $b->{$key} ) { return 1 } # not equal
				}
			}
		}
	}
	
	return 0; #equal
}

sub print_result {
	
	print_header(\@journals);

	foreach my $testcase ( @$order ) {
		my $line = $hash->{$testcase};
		
		foreach (@$line) { $_ = undef if defined $_ && $_ eq "?" };
		
		# Check if there are differences for this $testcase
		my $differs = 0;
		my $last = $line->[0];
		for ( my $i = 1; $i < $number; $i++ ) {
			if (   (defined $last && defined $line->[$i] || $cmp_undef)
				&& cmp_jentries( $line->[$i], $last ) != 0 )
			{
				$differs = 1;
				last;
			}
			
			$last = $line->[$i] if defined $line->[$i] || $cmp_undef;
		}
		
		if ( $differs || $show_equal ) {
			print_diff_line( $testcase, $line, $differs );
		}
	}

	print_footer();
}
#---------------

sub print_diff_line {
	my ($testcase, $line, $differs) = @_;
	
	$box_id++;
	
	print '<tr class="b" onclick="return sd(\''.$box_id.'\');">';
	
	print "<td>";
	# Print an invisible div with detailed info on differences.
	print '<div style="display:none" id="b'.$box_id.'">';
	print '<h3><b>'.($report_diff ? 'Test:' : 'Test Case:').'</b> '.html($testcase).'</h3>';
	print '<hr/>';
	for ( my $i = 0; $i < $number; $i++ ) {
		print '<div style="padding:.5em;margin:0;'
				.( ($i % 2) ? 'background:#F0F0F0;' : '').'">';
		print '<p><b>'.html($journal_names[$i]).':</b> <u>'.html($journals[$i]).'</u></p>';
		if ( defined $line->[$i] ) {
			my $obj = $line->[$i];
			my $severity = "";
			if ( ref($obj) eq "" ) {
				$severity = $obj;
				if ( $severity eq "" ) { $severity = "passed" }
			}
			else {
				( ref($obj) eq "HASH" ) or die "Wrong entry type in jstruct" if DEBUG;
				
				$severity = $obj->{SEVERITY};
			}
			print '<p><b>Verdict:</b> <span style="background:'.severity_color($severity).'">'
					.html($severity).'</span></p>';
			
			if ( ref($obj) ne "" ) {
				if ( defined $obj->{RESOLUTION} && $obj->{RESOLUTION} ne "" ) {
					print '<p><b>Resolution:</b><pre>'.html(pre_wrap($obj->{RESOLUTION})).'</pre></p>';
				}
				if ( defined $obj->{DESCRIPTION} && $obj->{DESCRIPTION} ne "" ) {
					print '<p><b>Description:</b><pre>'.html(pre_wrap($obj->{DESCRIPTION})).'</pre></p>';
				}
				if ( defined $obj->{COMMENT} && $obj->{COMMENT} ne "" ) {
					print '<p><b>Comment:</b><pre>'.html(pre_wrap($obj->{COMMENT})).'</pre></p>';
				}
				if ( defined $obj->{MESSAGE} && $obj->{MESSAGE} ne "" ) {
					print '<p><b>Messages from the test:</b><pre>'.html(pre_wrap($obj->{MESSAGE})).'</pre></p>';
				}
				if ( defined $obj->{PROBLEM_INFO} ) { # Problem info from problem DB
					my $pr_info = $obj->{PROBLEM_INFO};
					
					if ( defined $pr_info->{'comment'} && $pr_info->{'comment'} ne "" ) {
						print '<p><b>Problem info:</b><pre>'.html(pre_wrap($pr_info->{'comment'})).'</pre></p>';
					}
					if ( defined $pr_info->{'lsb_report_id'} ) {
						my $pr_id = $pr_info->{'lsb_report_id'};
						my $link = "https://www.linuxfoundation.org/lsbprs/pr/".$pr_id;
						print '<p>Reported to the LSB Committee as'
								.' <a href="'.$link.'" target="_blank">PR '.$pr_id.'</a>.</p>';
					}
				}
				# "LINE_NUM"
			}
		}
		else {
			print '<p><b>Skipped</b></p>';
		}
		print '</div>';
		print '<hr style="margin:0"/>';
	}
	print '</div>';
	
	# Visible part: test case name
	if ( $differs ) {
		print html($testcase);
	} else {
		print '<span style="color:gray">'.html($testcase).'</span>';
	}
	print "</td>\n";
	
	for ( my $i = 0; $i < $number; $i++ ) {
		my $severity = undef;
		if ( defined $line->[$i] ) {
			my $obj = $line->[$i];
			if ( ref($obj) eq "" ) {
				$severity = $obj;
				if ( $severity eq "" ) { $severity = "passed" }
			}
			else {
				( ref($obj) eq "HASH" ) or die "Wrong entry type in jstruct" if DEBUG;
				
				$severity = $obj->{SEVERITY};
			}
		}
		
		my $color = severity_color($severity);
		$color = " style='background-color:$color'" if ( $color );
		
		print "<td".$color.">\n";
		print $severity if defined $severity;
		print "</td>";
	}
	print "</tr>\n";
}

sub print_header {
	my ($journals) = @_;
	
	my $title = html( $report_diff ? 'Report Comparison' : 'Journal Comparison' );
	
	print <<HEREDOC;

<html>
<head>
	<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
	<title>$title</title>

<style type="text/css">
.b { cursor: pointer }
</style>

<script language="javascript" type="text/javascript">
// <![CDATA[
function sd (box_id) {
	var w = window.open( '', 'diffdetails', 'toolbar=0,menubar=0,height=750,width=850,dependent=yes,dialog=yes');
	w.moveTo(100,0);
	w.document.body.innerHTML = document.getElementById('b' + box_id).innerHTML;
	w.focus();
	return false;
}
// ]]>
</script>
</head>

HEREDOC
	
	print '<table border="1">'."\n";
	if ( defined $journals && @$journals ) {
		print "<tr>\n";
		print '<th>'.($report_diff ? 'Test:' : 'Test Case:').'</th>'."\n";
		
		my $i = 0;
		foreach my $jname ( @$journals ) {
			print '<th>';
			print '<a href="'.html($jname).'" title="'.html($jname).'" target="_blank">'
					.$journal_names[$i].' </a>';
			print "</th>\n";
		}
		continue { $i++; };
		
		print "</tr>\n";
	}
}

sub print_footer {
	print "</table>\n";
}

sub severity_color {
	my ($severity) = @_;
	
	if ( !defined $severity ) {
		return "";
	}
	
	if ( defined $Report::severity_map{$severity} ) {
		$severity = $Report::severity_map{$severity};
	}
	
	if ( $severity eq "" || $severity eq "passed" || $severity eq "untested" ) {
		return "#40FF40";
	}
	elsif ( $severity =~ /(waived|quelled)/ 
		|| $severity eq "fip" || $severity eq "notice" ) 
	{
		return "#C8C8C8";
	}
	elsif ( $severity =~ /(warning)/ ) {
		return "#FFFF40";
	}
	else {
		return "#FF6060";
	}
	
	return "";
}
