#!/usr/bin/perl -w

# LSB Test Execution Framework
# Main Executable Script (lsb-tef.pl)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use strict;
use IO::Handle; # for autoflush();
use FindBin;
use File::Basename qw(dirname basename);

use constant TEF_VERSION => "4.0.1";

sub BEGIN {
	# Add current directory to the @INC to be able to include modules
	# from the same directory where this script is located.
	unshift @INC, $FindBin::Bin;
	
	$| = 1; # Immediate flush of all output
	
	### LSB 3.1 / 16.4. Path For System Administration Utilities ##
	# Certain utilities used for system administration (and other privileged
	# commands) may be stored in /sbin, /usr/sbin, and /usr/local/sbin. 
	# Applications requiring to use commands identified as system 
	# administration utilities should add these directories to their PATH.
	$ENV{PATH} .= ":/sbin:/usr/sbin";
}

use Misc;
Misc::Init();
use Manifest;
use Packages;
use Download;
use UserProfile;
use Report;

# Test modules: test name -> test object.
our %tests = ();

# Array of tests to run; they will be run in the specified _order_. 
my @tests_to_run = ();

#----------------------------------------------------------------------------

$SIG{__WARN__} = sub {
		die "Error: Compilation error: ".$_[0] if !defined $^S; # Compilation error
		
		local $Error::Depth = $Error::Depth + 1;
		local $Error::Debug = 0;
		warning $_[0];
	};

$SIG{__DIE__} = sub {
		die @_ if $^S; # Do nothing if called inside eval{ }
		die "Error: Compilation error: ".$_[0] if !defined $^S; # Compilation error
		
		local $Error::Depth = $Error::Depth + 1;
		local $Error::Debug = 0;
		fail "Died. ".$_[0];
	};

my $need_cleanup = 0;

my $_END = sub {
	print STDERR "Finished.\n" if $globals->{'webui'};
};

#=============================================================================
# Subroutines of the main script
#=============================================================================

# Prints a short help text.
sub print_help {
	my ($manifest) = @_;
	
	# Get available tests
	my @test_names = sort grep !$manifest->{MANUAL}{$_}, @{$manifest->{ALL}{TESTSUITES}};
	
	# Pretty print the test names
	my $test_names = "   ";
	my $max_len = 1;
	map { $max_len = $_ if $_ > $max_len } map length, @test_names; # Select the max length
	my $columns = int(80/$max_len) - 1;
	for ( my $i = 1; $i <= @test_names; $i++ ) {
		my $test_name=$test_names[$i-1];
		$test_names .= $test_name;
		if ( $i % $columns == 0 ) {
			$test_names .= "\n   ";
		} else {
			$test_names .= " " x ($max_len + 1 - length $test_name);
		}
	}
	
	my $text = <<DATA;
 Usage: $0 [OPTIONS] TEST_NAMES
 Runs one or more tests and produces HTML report.

 Supported automated tests:
$test_names
 Use 'all' to run all available test.

 Options:
  
  --arch <architecture>   Specify the machine architecture. [autodetect]
  --cert                  Certification mode.
  --check-only            Make some checks and exit.
  -D,--download           Allow downloading needed files from the Internet.
                          Use -Dftp to use the FTP protocol instead of the HTTP.
  --force-reinstall       Reinstall already installed packages.
  -h,--help               Show this help and exit.
  -iu,--ignore-unavailable  Don't fail if unable to run one of the tests required.
  -s,--standard=LSB4.0    Specify the standard version. [autodetect]
  -p,--std-profile 'core,c++'  Specify the standard profile.
  --list                  Print all available test modules and their statuses.
  -f,--profile <file>     Take all settings from the file specified.
  --report <result dir>   Make report in the specified result directory.
  --keep                  Don't rebuild existing report parts.
  --status <status>       Choose test versions using this status (e.g. 'beta').
  -r,--testrun-id <name>  Specify the name of the result subdirectory.
  -v,--verbose <N>        Verbose mode (0=silent, 1=normal, 2=verbose, 3=debug).

 WARNING! PLEASE DO NOT USE THIS SCRIPT ON ANY MACHINE THAT HOLDS
 IMPORTANT DATA, OR PROVIDES IMPORTANT SERVICES!
 The tests sometimes put a heavy stress on the tested computer,
 and in rare unfortunate cases they can crash the system or damage
 the stored data. They also are a security risk as they work with
 root priviledges and create users with default passwords.
 
 Please report any bugs and improvements to linux\@ispras.ru.
DATA
	
	print $text;
}
#----------------------------------------------------------------------------

# Read the command line parameters.
sub get_options {
	my @args = @ARGV;
	if ( @args < 1 ) {
		print "Read $0 --help for information about usage.\n";
		exit 0;
	}
	
	my $params = [
			{LONG_NAME=>'arch',               SHORT_NAME=>'',	HAS_VALUE=>1},
			{LONG_NAME=>'cert'},
			{LONG_NAME=>'check-only'},
			{LONG_NAME=>'download',           SHORT_NAME=>'D'},
			{LONG_NAME=>'force-reinstall'},
			{LONG_NAME=>'help',               SHORT_NAME=>'h'},
			{LONG_NAME=>'ignore-unavailable', SHORT_NAME=>'iu'},
			{LONG_NAME=>'keep'},  # Don't rebuild existing report parts
			{LONG_NAME=>'standard',           SHORT_NAME=>'s',	HAS_VALUE=>1},
			{LONG_NAME=>'std-profile',        SHORT_NAME=>'p',	HAS_VALUE=>1},
			{LONG_NAME=>'list'},
			{LONG_NAME=>'mail-to',            SHORT_NAME=>'m',	HAS_VALUE=>1},
			{LONG_NAME=>'package-manager',    SHORT_NAME=>'pm',	HAS_VALUE=>1},
			{LONG_NAME=>'profile',            SHORT_NAME=>'f',	HAS_VALUE=>1},
			{LONG_NAME=>'refresh-prdb'},
			{LONG_NAME=>'report',             SHORT_NAME=>'',	HAS_VALUE=>1},
			{LONG_NAME=>'status',             SHORT_NAME=>'',	HAS_VALUE=>1},
			{LONG_NAME=>'testrun-id',         SHORT_NAME=>'r',	HAS_VALUE=>1},
			{LONG_NAME=>'verbose',            SHORT_NAME=>'v',	HAS_VALUE=>1},
			{LONG_NAME=>'webui',              SHORT_NAME=>'w'},
		];
	
	my $options = {};
	
	while ( @args ) {
		my $arg = shift @args;
		
		if ( $arg =~ /^-/ ) {
			my $found = 0;
			foreach my $opt ( @$params ) {
				my $long_name = $opt->{'LONG_NAME'};
				my $short_name = $opt->{'SHORT_NAME'};
				my $value = undef;
				
				if ( $long_name && $arg =~ m/^--\Q$long_name\E=(.+)$/ ) {
					$value = $1;
				}
				elsif ( $short_name && $arg =~ m/^-\Q$short_name\E(.+)$/ ) {
					$value = $1;
				}
				elsif ( $long_name && $arg =~ m/^--\Q$long_name\E$/ ) {
					# OK
				}
				elsif ( $short_name && $arg =~ m/^-\Q$short_name\E$/ ) {
					# OK
				}
				else {
					next;
				}
				
				if ( !defined $value ) {
					if ( $opt->{HAS_VALUE} ) {
						$value = shift @args;
						( defined $value ) or return error "Option $arg must be followed by a value";
					} else {
						$value = 1;
					}
				}
				
				my $result_name = ($long_name or $short_name);
				
				$options->{$result_name} = $value;
				$found = 1;
				last;
			}
			if ( !$found ) {
				fail "Unexpected parameter: '$arg'";
			}
		} else {
			# this is not a parameter, so it's probably a test name
			push @{$options->{TESTS}}, $arg;
		}
	}
	return $options;
}
#----------------------------------------------------------------------------

sub Setup {
	# Some global initializations
	$globals->{'start_time'} = time();
	# Process ID
	$globals->{'PID'} = $$;
	# Script directory
	$globals->{'script_directory'} = $FindBin::Bin;
	# Current profile
	$globals->{'std_profile'} = $profile_desktop; # default profile
	# File with the current run status
	$globals->{'status_file'} = RESULTS_DIR.'/'.STATUS_FILE;
	# Ptyshell tool location
	$globals->{'ptyshell'} = $FindBin::Bin."/../ptyshell/ptyshell";
	
	# Parse command line args
	my $options = get_options();
	is_ok($options) or fail "Error in command line parameters", $Error::Last;
	
	# Read the profile before processing other options
	if ( $options->{'profile'} ) {
		my $profile_filename = $options->{'profile'};
		
		if ( !defined $profile_filename || !-f $profile_filename ) {
			fail "Wrong profile filename "
					.(defined $profile_filename ? "'$profile_filename'" : "(undefined)");
		}
		$globals->{'profile_filename'} = $profile_filename;
		
		$globals->{'profile'} = read_profile($profile_filename)
			or fail $UserProfile::profile_error;
	
		is_ok Use_profile_general( $globals->{'profile'}, $options )
			or fail "Failed to use profile '$profile_filename'", $Error::Last;
	}
	
	# --
	
	if ( $options->{'verbose'} ) {
		$Misc::verbose_lvl = int($options->{'verbose'});
	}
	if ( $options->{'standard'} ) {
		my $standard = $options->{'standard'};
		if ( $standard =~ /^(\w+)\s*(\d\S*)/ ) {
			$globals->{'standard'} = "$1 $2";
		} else {
			fail "Wrong standard name. Expected something like 'Moblin2.0' or 'LSB4.0'.";
		}
	}
	if ( $options->{'std-profile'} ) {
		$globals->{'std_profile'} = $options->{'std-profile'};
		
		in_array( $globals->{'std_profile'}, @all_profiles )
			or warning "Wrong standard profile: ".$globals->{'std_profile'};
	}
	if ( $options->{'arch'} ) {
		my $arch = $options->{'arch'};
		$host_info->{'architecture'} = guess_architecture($arch)
			or fail "Wrong architecture: '$arch'."
				." Allowed values are: ".(join ", ", @all_archs).".";
		$host_info->{'machine'} = $arch;
	}
	if ( $options->{'package-manager'} ) {
		my $package_manager = lc $options->{'package-manager'};
		( $package_manager eq 'rpm' || $package_manager eq 'dpkg' )
			or fail "Wrong package manager: '$package_manager'."
				." Supported package managers are 'rpm' and 'dpkg' (using 'alien').";
		$host_info->{'package_manager'} = $package_manager;
	}
	if ( $options->{'status'} ) {
		my $status = $options->{'status'};
		
		( in_array( $status, @all_statuses ) )
			or fail "Wrong status: '$status'. Allowed values are: @all_statuses.";
		
		$globals->{'status'} = $status;
	} else {
		$globals->{'status'} = 'cert'; # Default status
	}
	if ( $options->{'download'} ) {
		$Download::enabled = $options->{'download'};
		if ( $options->{'download'} ne 1 ) {
			foreach my $s ( split /,/, $options->{'download'} ) {
				if ( $s =~ /^(http|ftp)$/i ) {
					$globals->{'dl_method'} = lc $1; # Used in Manifest::file_link();
				}
				if ( $s =~ /^(wget|curl)$/i ) {
					$Download::tool = lc $1;
				}
				else {
					fail "Can't understand the -D parameter. It should be -D[ftp|http|wget|curl] or comma-separated: -Dftp,wget.";
				}
			}
		}
	}
	if ( $options->{'mail-to'} ) {
		$globals->{'mailto'} = $options->{'mail-to'};
	}
	if ( $options->{'force-reinstall'} ) {
		$globals->{'force_reinstall'} = $options->{'force-reinstall'};
	}
	if ( $options->{'check-only'} ) {
		$globals->{'check_only'} = $options->{'check-only'};
	}
	if ( $options->{'testrun-id'} ) {
		$globals->{'testrun_id'} = $options->{'testrun-id'};
	}
	if ( $options->{'webui'} ) {
		$globals->{'webui'} = $options->{'webui'};
	}
	
	# ---
	
	if ( $options->{'help'} ) {
		if ( !$options->{'verbose'} ) { # Not specified explicitly
			$Misc::verbose_lvl = 0;
		}
		debug_inform "Loading manifest...";
		my $manifest = load_manifest();
		is_ok $manifest or fail "Failed to load manifest.", $Error::Last;
		print_help($manifest);
		exit 0;
	}
	if ( $options->{'refresh-prdb'} ) {
		if ( $options->{'refresh-prdb'} ne "1" ) {
			$Download::enabled = 1;
			host_check_and_info();
			is_ok refresh_prdb() or fail $Error::Last;
			exit 0;
		}
		# else:
		$globals->{'refresh_prdb'} = 1;
	}
	if ( $options->{'report'} ) {
		my $dir = $options->{'report'};
		if ( $dir !~ m{/} ) { $dir = RESULTS_DIR."/".$dir; }
		
		# Complain if not root
		$> == 0 or fail "Please run this script as root.\n";
		
		rebuild_report($dir, $options->{'keep'});
		
		exit 0;
	}
	
	if ( $options->{'list'} ) {
		inform "Loading test modules...";
		my $manifest = $globals->{'manifest'} = load_manifest();
		is_ok($manifest) or fail "Failed to load manifest.", $Error::Last;
		
		host_check_and_info(); # get the host info
		$Packages::package_manager = $host_info->{'package_manager'};
	
		list_available_versions();
		exit 0;
	}
	
	# ---
	
	# If reached this point, there are some tests to be run.
	
	# Complain if not root
	$> == 0 or fail "Please run this script as root.\n";
	
	host_check_and_info(); # get the host info
	$Packages::package_manager = $host_info->{'package_manager'};
	
	## Check the PtyShell tool
	if ( !-x $globals->{'ptyshell'} ) {
		return error "The ptyshell utility (".$globals->{'ptyshell'}.")"
				." doesn't exists or isn't executable.";
	}
	
	# Initialize logging
	inform "Initialization...";
	
	if ( !defined $globals->{'testrun_id'} ) {
		my ($sec, $min, $hour, $mday, $mon, $year) = localtime( $globals->{'start_time'} );
		$globals->{'testrun_id'} = 
			 ( defined $host_info->{'architecture'} ? $host_info->{'architecture'} : "" )
			.'-'.(defined $host_info->{'hostname'} ? $host_info->{'hostname'} : "" )
			.'-'.sprintf( '%04d-%02d-%02d-%02dh-%02dm-%02ds',
					$year + 1900, $mon + 1, $mday, $hour, $min, $sec );
	}
	
	# Create needed dirs and initialize logging
	is_ok setup_dirs_and_logs() or return $Error::Last;
	
	inform "... initialization done.";
	inform "Result dir: '".$globals->{'result_dir'}."'" if $globals->{'result_dir'}; # The first line in log
	
	# Print some settings
	inform "==========================================";
	inform "LSB Test Execution Framework v. ".TEF_VERSION;
	inform format_time( $globals->{'start_time'} );
	inform "Command line: @ARGV" if "@ARGV" ne "";
	inform "Script directory is $FindBin::Bin";

	inform "Host name: ".$host_info->{'hostname'};
	inform "Host machine: ".$host_info->{'machine'};
	inform "Host architecture: ".$host_info->{'architecture'};
	inform "Host kernel: ".$host_info->{'kernel'};
	inform "Host OS: ".$host_info->{'OS'};
	inform "Host package manager: ".$host_info->{'package_manager'};
	inform "Standard: ".$globals->{'standard'};
	inform "==========================================";
	
	if ( defined $globals->{'profile_filename'} ) {
		cmd( "cp -f ".shq($globals->{'profile_filename'})." ".shq($globals->{'result_dir'}."/profile") ) == 0
			or warning "Failed to copy profile: '".$globals->{'profile_filename'}."' to '".$globals->{'result_dir'}."'";
	}
	
	# Load the Manifest
	inform "Loading manifest...";
	my $manifest = $globals->{'manifest'} = load_manifest();
	is_ok $manifest or fail "Failed to load manifest.", $Error::Last;
	inform "...done.";
	
	# The --cert option should be checked at the very end, because it overrides many parameters
	if ( $options->{'cert'} ) {
		$globals->{'cert_mode'} = 1;
		
		$options->{'ignore-unavailable'} = 0;
		$globals->{'status'} = 'cert';
		$options->{TESTS} = ['all', 'manual'];
		
		$globals->{'force_reinstall'} = 1;
		
		# TODO: refresh manifest
		# TODO: refresh problem_db
	}

	# Choose tests version
	inform "Choosing tests version...";
	debug_inform("Status: '".$globals->{'status'}."'");
	
	@tests_to_run = ();
	
	my $standard = $globals->{'standard'};
	my $arch = $host_info->{'architecture'};
	my @add = @{$options->{TESTS}} if $options->{TESTS};
	while ( @add ) {
		my $test_name = shift @add;
		
		# Alias 'all'
		if ( $test_name eq 'all' ) {
			for my $testsuite ( sort keys %{$manifest->{TREE}{$standard}} ) {
				next if $manifest->{MANUAL}{$testsuite}; # All but manual
				next if !glob_hash($manifest, 'COMPOSITION', $standard, $testsuite, $arch, $globals->{'status'}); # Skip unavailable
				push @add, $testsuite;
			}
			next;
		}
		# Alias 'manual'
		if ( $test_name eq 'manual' ) {
			for my $testsuite ( sort keys %{$manifest->{MANUAL}} ) {
				next if !glob_hash($manifest, 'COMPOSITION', $standard, $testsuite, $arch, $globals->{'status'}); # Skip unavailable
				push @add, $testsuite;
			}
			next;
		}
		# Check if the test is specified in the Manifest
		if ( !in_array( $test_name, @{$manifest->{ALL}{TESTSUITES}} ) ) {
			my $error = "Test '$test_name' is unknown."
					."Use the --list option to get the list of available tests.";
			$options->{'ignore-unavailable'} ? 	warning $error : fail $error;
		}
		
		# Init test info according to Manifest
		my $test_obj = init_test_info( $test_name );
		is_ok($test_obj) or fail $Error::Last;
		
		# Select the test version, files for installation and set up the test options
		my $res = Choose_test_version( $test_obj );
		if ( !is_ok($res) || !defined $test_obj->{VERSION} ) {
			my $error = error "Test '$test_name' can't be run with the current settings.", $Error::Last;
			$options->{'ignore-unavailable'} ? 	warning $error : fail $error;
			next;
		}
		
		# Load test module and create test object
		my $module_file = prepare_modules( $test_obj );
		if ( !is_ok($module_file) ) {
			my $error = error "Failed to prepare modules for '$test_name'.", $Error::Last;
			$options->{'ignore-unavailable'} ? 	warning $error : fail $error;
			next;
		}
		my $class_name = load_test_module($module_file); # TODO: load all test modules for debug purposes.
		if ( !is_ok($class_name) ) {
			my $error = error "Failed to load module '$module_file'.", $Error::Last;
			$options->{'ignore-unavailable'} ? 	warning $error : fail $error;
			next;
		}
		$test_obj = $class_name->new( %$test_obj ); # create test object
		
		$tests{$test_name} = $test_obj;
		
		push @tests_to_run, $test_name;
	}
	
	if ( !@tests_to_run ) {
		fail "No tests to run (use 'all' to run all available tests).";
		# TODO: don't create directories until this point to avoid appearing empty results.
	}
	
	# Print the tests to be run with required files.
	inform "Tests to be run:";
	foreach my $test_name ( @tests_to_run ) {
		my $test_obj = $tests{$test_name};
		( defined $test_obj->{VERSION} ) or warning("Version undefined for '$test_name'"), next;
		
		inform "   $test_name ".$test_obj->{VERSION};

		# Print more detailed information about files to be installed
		( defined $test_obj->{FILES} && @{$test_obj->{FILES}} )
			or warning("No files selected for '$test_name'"), next;
		
		foreach ( @{$test_obj->{FILES}} ) {
			debug_inform "      : ".$_->{FILENAME};
		}
	}
	
	$globals->{'verdict'} = 'incomplete';
	$globals->{'verdict_man'} = 'incomplete';
	
	return 1;
}
#----------------------------------------------------------------------

# Checks whether the specified tool is available, terminating immediately
# if it is not.
sub require_tool {
	my ($tool_name) = @_;
	
	# Look if the shell knows this tool.
	unless ( does_shell_know($tool_name) ) {
		fail "Couldn't find a required tool: '$tool_name'."
		   ."\nYou could install it using the tools provided by your distro.";
	}
}
#----------------------------------------------------------------------

# This function creates an object that works with the test suite.
sub load_test_module {
	my ($file) = @_;
	
	debug_inform "Loading module: $file";
	
	# Link up the module and create a test object.
	my $test_obj;
	eval {
		my $dir = extract_dir( $file );
		push @INC, $dir if !in_array($dir, @INC);
		require $file;
	};
	if ($@) { # Any errors?
		return error "Failed to load module: '$file'", error $@;
	}
	
	my $class_name = extract_filename( $file );
	$class_name =~ s/\.pm$//;
	
	return $class_name;
}
#-----------------------------------------------------------------------
sub init_test_info {
	my ($test_name) = @_;
	
	my $manifest = $globals->{'manifest'} or return error "Manifest wasn't loaded";
	my $standard = $globals->{'standard'};
	
	my $test_obj = { NAME => $test_name };
	
	$test_obj->{TREE} = $manifest->{TREE}{$standard}{$test_name}
		or return error "Test '$test_name' is unavailable for $standard",
				error "Tree entry isn't specified in the Manifest "
					." for test '$test_name', $standard.";
	
	# ->{DESCRIPTIONS}{$test_name}[]->{STANDARD|DISPLAYNAME|DESCRIPTION} = ".."
	my ($descr) = grep {version_filter($_->{STANDARD}, $standard)} @{$manifest->{'DESCRIPTIONS'}{$test_name}};
	if ( !$descr ) {
		return error "There is no description in the Manifest"
					." for test '$test_name', $standard.";
	}
	
	# Description & position in the TOC tree
	( $descr->{DISPLAYNAME} ) or warning "DISPLAYNAME for '$test_name' is not defined";
	$test_obj->{DISPLAYNAME} = $descr->{DISPLAYNAME};
	
	( $descr->{DESCRIPTION} ) or warning "DESCRIPTION for '$test_name' is not defined";
	$test_obj->{DESCRIPTION} = $descr->{DESCRIPTION};
	
	return $test_obj;
}
#-----------------------------------------------------------------------

sub prepare_modules {
	my ($test_obj, $pick) = @_;
	
	my $manifest = $globals->{'manifest'};
	( $test_obj->{VERSION} ) or return error "Test version isn't defined";
	my $test_name = $test_obj->{NAME};
	
	my @modules = grep {
			   version_filter($_->{TEST_VER}, $test_obj->{VERSION})
			&& version_filter($_->{STANDARD}, $globals->{'standard'})
		} @{$manifest->{MODULES}{$test_name}};
	
	my %found_files = ();
	for my $module ( @modules ) {
		my $module_file = $module->{FILE};
		my $fullname = $globals->{'script_directory'}."/Tests/".$module_file;
		
		# Update the module
		my $from = $module->{FROM};
		my $err = undef;
		if ( $from && $pick ) {
			if ( $from =~ /^\// ) {
				if ( -f $from ) {
					cmd("cp -f ".shq($from)." ".shq($fullname));
					( -f $fullname ) or $err = error $last_cmd_output;
				} else {
					$err = error "There is no file '$from'";
				}
			}
			elsif ( $from =~ /^(https?|ftps?):/ ) {
				is_ok download_file($from, $fullname)
					or $err = error "Failed to download '$from'", $Error::Last;
			}
		}
		
		if ( !-f $fullname ) {
			return error "Can't find module file '$module_file'", $err;
		} else {
			if ( !is_ok($err) ) { warning $err; warning "Using old module '$module_file'"; }
		}
		
		$found_files{$module_file} = $fullname;
	}
	
	my ($main_module) = grep $_->{MAIN}, @modules;
	( $main_module ) or return error "No module specifed in Manifest for '$test_name' ver ".$test_obj->{VERSION}.", ".$globals->{'standard'};
	
	return $found_files{$main_module->{FILE}};
}
#-----------------------------------------------------------------------

# Select test version, files for installation and set up the test options
sub Choose_test_version {
	my ($test_obj) = @_;
	
	my $test_name = $test_obj->{NAME};
	
	my $architecture = $host_info->{'architecture'};
	my $manifest = $globals->{'manifest'};
	my $profile = $globals->{'profile'};
	my $standard = $globals->{'standard'};
	
	my $profile_entry = undef;
	$profile_entry = $profile->{TEST_SUITES}{$test_name} if defined $profile;
	if ( defined $profile_entry ) { # User profile
		$test_obj->{VERSION} = $profile_entry->{VERSION};
		$test_obj->{STATUS}  = $profile_entry->{STATUS};
	}
	
	if ( !defined $test_obj->{STATUS} ) {
		$test_obj->{STATUS} = $globals->{'status'};
	}
	
	if ( !defined $test_obj->{VERSION} ) {
		my $status = $test_obj->{STATUS};
		$test_obj->{VERSION} = select_test_version_by_status( $manifest,
					$status, $test_name, $standard, $architecture );
		
		if ( !defined $test_obj->{VERSION} ) {
			return error "$test_name test isn't supported for"
				." $standard, arch '$architecture', status '$status'.";
		}
	}
	
	# $profile->{FILES}->{<file>}->{DOWNLOAD} == <dl-option>
	#                            ->{INSTALL} == <inst-option>
	# <inst-option> is one of the following:
	#   0 - install, if necessary;  1 - force reinstall;  2 - do not deinstall
	# <dl-option> is one of the following:
	#   0 - download, if necessary;  1 - force download;   2 - do not download
	if ( defined $profile && defined $profile->{FILES} ) {
		
		foreach my $key ( keys %{$profile->{FILES}} ) {
			my $file = $manifest->{FILES}{$key};
			
			if ( !defined $file ) {
				warning "File '$file' isn't defined in Manifest";
				next;
			}
			
			$file->{OPTIONS} = $profile->{FILES}{$key};
			
			my $package_name = $file->{PACKAGE};
			my $package_ver = $file->{VERSION};
			if ( defined $package_name && defined $package_ver ) {
				if ( defined $file->{OPTIONS}{INSTALL} && $file->{OPTIONS}{INSTALL} == 2 ) {
					# <inst-option> = 2 - do not deinstall, keep installed
					$manifest->{PACKAGES}{$package_name}{KEEP}{$package_ver} = 1;
				}
			}
		}
	}
	###
	
	# Select files
	my $files = select_files( $manifest, $test_name, $test_obj->{VERSION}, $test_obj->{STATUS}, $standard, $architecture );
	if ( is_err($files) || !$files || !@$files ) {
		return error "Couldn't select files for '$test_name'.", $Error::Last;
	}
	$test_obj->{FILES} = $files;
	
	# Options & questions
	if ( defined $manifest->{OPTIONS}{$test_name} ) {
		my $version = $test_obj->{VERSION};
		my @options = grep { version_filter($_->{TEST_VER}, $version) 
				&& version_filter($_->{STANDARD}, $standard) } @{$manifest->{OPTIONS}{$test_name}};
		
		foreach my $opt ( @options ) {
			my $ID = $opt->{ID};
			
			( !defined $test_obj->{OPTIONS}{$ID} )
				or warning "Double option: '$ID'";
			
			# Copy the option with (a default value) from the Manifest
			$test_obj->{OPTIONS}{$ID} = copy( $opt );
		}
		
		# Get option values from the profile
		if ( defined $profile_entry ) {
			my $values = $profile_entry->{OPTIONS};
			if ( $values && %$values ) {
				foreach my $ID ( keys %$values ) {
					if ( !defined $test_obj->{OPTIONS}{$ID} ) {
						warning "Unused option for '$test_name': '$ID'";
						next;
					}
					$test_obj->{OPTIONS}{$ID}->{VALUE} = $values->{$ID};
				}
			}
			
			$values = $profile_entry->{AUTO_REPLIES}->{$test_obj->{STATUS}}->{$test_obj->{VERSION}};
			if ( $values ) {
				foreach my $ID ( keys %$values ) {
					if ( !defined $test_obj->{OPTIONS}{$ID} ) {
						warning "Unused option for '$test_name': '$ID'";
						next;
					}
					$test_obj->{OPTIONS}{$ID}->{VALUE} = $values->{$ID};
				}
			}
		}
	} # end of  if ( defined $manifest->{OPTIONS}{$test_name} )
	
	return 1;
}
#----------------------------------------------------------------------

sub select_test_version_by_status {
	my ($manifest, $status, $test_name, $standard, $architecture) = @_;
	
	my $test_obj = $tests{$test_name};
	
	# Manifest structure:
	# ->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}->{$test_ver}[]->{PACKAGE|PACKAGE_VER|LOCATION} = ".."
	
	my $test_vers = $manifest->{COMPOSITION}{$standard}{$test_name}{$architecture}{$status};
	
	if ( !$test_vers ) {
		return undef;
	}
	
	my $version_max = (sort { vercmp($a,$b) } keys %$test_vers)[-1];
	
	return $version_max;
}
#----------------------------------------------------------------------

# Choose which packages should be installed 
#  for selected test, version, standard, architecture.
sub select_files {
	my ($manifest, $testsuite, $test_ver, $status, $standard, $arch) = @_;
	
	my $package_manager = $host_info->{'package_manager'};
	
	my @result = ();
	
	my $requirements = $manifest->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}{$test_ver}
		or return error "Requirements aren't specified for"
					." '$standard' - '$testsuite' '$test_ver', '$arch', status '$status';";
	
	foreach my $req ( @$requirements ) {
		my $package_name = $req->{PACKAGE};
		my $package_ver  = ( $req->{PACKAGE_VER} or "*" );
		my $location     = ( $req->{LOCATION} or "" );
		
		my $found_files = Manifest::select_file( $manifest, $arch, $package_manager,
					$package_name, $package_ver, $location );
		
		if ( is_err($found_files) || !$found_files || !@$found_files ) {
			return error "select_files: Can't find file '$package_name' v. '$package_ver'"
				." ($package_manager), '$standard', arch: '$arch',"
				." required by $testsuite".($location ? " in '$location'": ""),
				$Error::Last;
		}
		if ( @$found_files > 1 ) {
			warning "Ambigous file choise: ".join ", ", map {$_->{SRC}} @$found_files;
		}
		
		my $file = $found_files->[0]; # Take any
		
		push @result, $file;
	}
	
	return \@result; # OK
}
#----------------------------------------------------------------------

# Do initial environment checks and read information about the host machine.
# Terminates the script if any of the checks fails.

sub host_check_and_info {
	# Check for basic utilities we will need.
	require_tool("uname"); # Included since LSB 1.0
	
	if ( !defined $host_info->{'OS'} ) {
		$host_info->{'OS'} = detect_OS();
	}
	
	if ( !defined $globals->{'standard'} ) {
		$globals->{'standard'} = detect_standard_version($host_info->{'OS'});
	}
	
	if ( $globals->{'standard'} =~ /^LSB/ ) {
		require_tool("lsb_release"); # This tool is required by the LSB
		
		if ( capture ("lsb_release 2>&1") =~ /No LSB modules/ ) {
			fail "No LSB modules are available. Expected that lsb_release prints available LSB modules.";
		}
	}

	if ( $Download::enabled ) { 
		if ( $Download::tool eq 'curl' ) {
			if( !does_shell_know("curl") ) {
				if( !does_shell_know("wget") ) { # try wget instead of curl
					fail ( "Couldn't find neither 'curl' nor 'wget'." );
				} else {
					inform "'curl' is unavailable, using 'wget' instead.";
					$Download::tool = 'wget';
				}
			}
		}
		elsif ( $Download::tool eq 'wget' ) {
			if( !does_shell_know("wget") ) {
				if( !does_shell_know("curl") ) { # try curl instead of wget
					fail ( "Couldn't find neither 'curl' nor 'wget'." );
				} else {
					inform "'wget' is unavailable, using 'curl'.";
					$Download::tool = 'curl';
				}
			}
		}
	}
	
	if ( $globals->{'mailto'} ) {
		does_shell_know("/usr/sbin/sendmail")
			or fail "Couldn't find '/usr/sbin/sendmail'. This tool is used for sending results by e-mail.";
	}
	
	# TODO: upload to a remote host
	#if ( $rlocation ) {
		#require_tool("expect");
		#require_tool("rsync");
	#}
	
	my $tmp = "";
	
	if ( !defined $host_info->{'machine'} ) {
		# The machine type and architecture are discovered via uname.
		$tmp = `uname -m`; chomp $tmp;
		if ( $tmp ne "" ) {
			$host_info->{'machine'} = $tmp;
		} else {
			fail "Failed to recognize the machine type."
				."\nUse the --arch option to specify the machine type manually on the command line";
		}
	}
	
	if ( !defined $host_info->{'hostname'} ) {
		# Get the host name.
		$tmp = `uname -n`; chomp $tmp;
		if ( $tmp ne "" ) {
			$host_info->{'hostname'} = $tmp;
		} else {
			fail "Failed to obtain the host name."
					."\n'uname -n' command should print the host name.";
		}
	}
	
	if ( !defined $host_info->{'architecture'} ) {
		$host_info->{'architecture'} = detect_architecture()
			or fail "Failed to recognize the machine architecture."
				."\nSupported architectures are ".(join ", ", @all_archs)."."
				."\nUse the --arch option to specify the architecture manually on the command line"
				."\nYou can identify the architecture using the command 'uname -m'.";
	}
	
	if ( !defined $host_info->{'kernel'} ) {
		# Get the kernel version
		$tmp = `uname -r -v`; chomp $tmp;
		$host_info->{'kernel'} = $tmp;
	}
	
	# Read the /etc/issue file, if present. Sometimes it is the only place
	# where an exact OS version can be found.
	if ( !defined $host_info->{'etc_issue'} && -f "/etc/issue" ) {
		$tmp = read_file("/etc/issue");
		if ( is_ok($tmp) ) {
			$tmp =~ s/\s+$//; # Remove trailing spaces
			$host_info->{'etc_issue'} = $tmp;
		}
	}
	
	if ( !defined $host_info->{'package_manager'} ) {
		$host_info->{'package_manager'} = guess_package_manager( $host_info->{'OS'} )
			or fail "Couldn’t recognize the package manager"
					." for this distro (".$host_info->{'OS'}.")."
					."\nUse the -pm option to specify the package manager manually on the command line"
					."\nSupported package managers are 'rpm' and 'dpkg' (using 'alien').";
	}
	
	# rpm is required even if the package manager is dpkg
	require_tool ("rpm");
	if ( $host_info->{'package_manager'} eq "dpkg") {
		# Check for alien and dpkg
		require_tool("alien"); # Alien is used to install rpm packages
		require_tool("dpkg");
	}
}
#----------------------------------------------------------------------

sub setup_dirs_and_logs {
	
	# Default temporary directory
	$globals->{'temp_dir'} = "/tmp/dtk-manager";
	is_ok create_empty_directory( $globals->{'temp_dir'} )
		or return error "Failed to create temp dir", $Error::Last;
	
	return 1 if $globals->{'check_only'};
	
	if ( ! -d RESULTS_DIR ) {
		is_ok create_empty_directory( RESULTS_DIR )
			or return error "Failed to create result dir: ".RESULTS_DIR;
	}
	
	# Results subdir
	$globals->{'result_dir'} = RESULTS_DIR."/".$globals->{'testrun_id'};
	if ( ! -d $globals->{'result_dir'} ) {
		# Create a subdir in the result directory
		mkdir $globals->{'result_dir'}
			or return error "Failed to create dir '".$globals->{'result_dir'}."': $!";
		is_ok append_string_to_file( '!'.RESULTS_DIR."/HISTORY", $globals->{'testrun_id'}."\n" )
			or return error "Failed to write to the HISTORY file", $Error::Last;
	}
	$Report::result_dir = $globals->{'result_dir'};
	
	$globals->{'test_result_dir'} = $globals->{'result_dir'}."/results";
	if ( ! -d $globals->{'test_result_dir'} ) {
		# Create directory in the result repository
		mkdir $globals->{'test_result_dir'}
			or return error "Failed to create dir '".$globals->{'test_result_dir'}."': $!";
	}
	
	## Setup logs

	## Output log
	my $log_fh;
	my $log_file = $globals->{'result_dir'}."/log";
	open $log_fh, ">>$log_file"
		or return error "Failed to open file '$log_file' for writing: $!";
	# Setup immediate flushing for log file handle.
	$log_fh->autoflush(1);
	$globals->{'log_file_handle'} = $log_fh;
	$globals->{LINE_N}{0} = 1;
	
	## Verbose log
	my $vlog_fh;
	my $vlog_file = $globals->{'result_dir'}."/verbose_log";
	open $vlog_fh, ">>$vlog_file"
		or return error "Failed to open log file '$vlog_file' for writing: $!";
	# Setup immediate flushing for log file handle.
	$vlog_fh->autoflush(1);
	$globals->{'vlog_file_handle'} = $vlog_fh;
	$globals->{LINE_N}{1} = 1;
	
	unless ( $globals->{'webui'} ) {
		open STDERR, "| tee -a '$vlog_file'" # Make STDERR be logged
			or return error "Failed to dup STDERR: $!";
		# TODO: use STDERR handle hook
	}
}
#----------------------------------------------------------------------

sub Benchmark {
	
	my $i;
	my $res;

	$SIG{ALRM} = sub { $res = $i; }; # Used to stop benchmarking after N second (alarm function)

	# Benchmark calculations
	my $r = 20;
	$i = 1; $res = undef; alarm 1;
	while ( !defined $res ) {  $i++;
		$r = length($r*$i*$i*$i); # arithmetic operations
	}
	$host_info->{'bench_c'} = $res/1000;
		
	# Benchmark process spawning
	$i = 1; $res = undef; alarm 2;
	while ( !defined $res ) {  $i++;
		`cat /etc/issue`; # spawn processes
	}
	$host_info->{'bench_p'} = $res;

	$SIG{ALRM} = 'DEFAULT';
	
	# Print benchmark results
	inform "Benchmark C: ".$host_info->{'bench_c'};
	inform "Benchmark P: ".$host_info->{'bench_p'};
	
	return 1;
}
#----------------------------------------------------------------------

sub write_runconfig {

	my $content = "";
	
	local *add_var = sub {
		my ( $name, $value ) = @_;
		
		# check
		if ( !defined $value ) {
			warning "$name value is undefined";
			$value = "";
		}
		
		$value =~ s{[\n\x0D]+}{\x81}g;
		
		$content .= $name.": ".$value."\n";
		
		return 1;
	};
	
	add_var( 'STANDARD', $globals->{'standard'} );
	add_var( 'CERT', $globals->{'cert_mode'} ) if $globals->{'cert_mode'};
	add_var( 'RUN_COMMENTS', $globals->{'run_comments'} ) if $globals->{'run_comments'};
	
	# Save the $host_info
	foreach my $key ( keys %$host_info ) {
		add_var( 'HOST_'.$key, $host_info->{$key} );
	}
	
	add_var( 'TESTER_NAME', $globals->{'tester_name'} ) if $globals->{'tester_name'};
	add_var( 'ORGANIZATION', $globals->{'tester_org'} ) if $globals->{'tester_org'};
	
	foreach my $test_name ( @tests_to_run ) {
		my $test_obj = $tests{$test_name};
		
		$content .= "\n[".$test_name."]\n";
		
		add_var( 'VERSION', $test_obj->{'VERSION'} );
		add_var( 'TREE', $test_obj->{'TREE'} );
		add_var( 'DISPLAYNAME', $test_obj->{'DISPLAYNAME'} );
		add_var( 'STATE', $test_obj->{'STATE'} );
		add_var( 'VERDICT', $test_obj->{'VERDICT'} ) if $test_obj->{'VERDICT'};
	}
	
	is_ok write_string_as_file( $globals->{'result_dir'}."/runconfig", $content )
		or return $Error::Last;
	
	return;
}

# Write the INFO file
sub write_INFO {
	my $INFO = "";
	
	$INFO .= "Verdict: ".Report::verdict_text($globals->{'verdict'})."\n" if $globals->{'verdict'};
	if ( grep { $tests{$_}{MANUAL} } @tests_to_run ) {
		$INFO .= "Verdict-manual: ".Report::verdict_text($globals->{'verdict_man'})."\n" if $globals->{'verdict_man'};
	}
	$INFO .= "Certification\n" if $globals->{'cert_mode'};
	
	$INFO .= $globals->{'standard'}." tests:\n" if $globals->{'standard'};
	
	foreach my $test_name ( @tests_to_run ) {
		my $test_obj = $tests{$test_name};
		next if $test_obj->{MANUAL};
		$INFO .= " $test_name";
		$INFO .= " - ".$test_obj->{VERDICT} if $test_obj->{VERDICT};
		$INFO .= ";\n";
	}
	if ( grep { $tests{$_}{MANUAL} } @tests_to_run ) {
		$INFO .= " Manual tests";
		$INFO .= " - ".$globals->{'verdict_man'} if $globals->{'verdict_man'};
		$INFO .= ";\n";
	}
	
	$INFO .= "\n";
	
	$INFO .= $globals->{'run_comments'} if $globals->{'run_comments'};
	
	is_ok write_string_as_file( $globals->{'result_dir'}."/INFO", $INFO )
		or return $Error::Last;
	
	return;
}
#-----------------------------------------------------------------------

# Write the status file. See TestStatus.pm for details.
sub write_status {
	my $content = "";
	
	$content .= "PID = ".$globals->{'PID'}."\n";
	$content .= "RESULT_DIR = ".$globals->{'testrun_id'}."\n" if defined $globals->{'testrun_id'}; # undef if --check-only
	$content .= "CERTIFICATION = ".($globals->{'cert_mode'} ? 1 : 0)."\n";
	$content .= "STATUS = ".($globals->{'done'} ? $globals->{'done'} : 'Running')."\n";
	
	$content .= "CURRENT_TIME = ".time()."\n";
	
	$content .= "START_TIME = ".$globals->{'start_time'}."\n";
	
	# Tests to run:
	foreach my $test_name ( @tests_to_run ) {
		my $test_obj = $tests{$test_name};
		
		$content .= "\n[".$test_name."]\n";
		
		$content .= "NAME = ".($test_obj->{DISPLAYNAME} or $test_name)."\n";
		
		$content .= "STATUS = ".$test_obj->{STATE}."\n";
		
		$content .= "START_TIME = ".$test_obj->{PREPARE_TIME}."\n" if defined $test_obj->{PREPARE_TIME};
		$content .= "STOP_TIME = ".$test_obj->{FINISH_TIME}."\n" if defined $test_obj->{FINISH_TIME};
		
		$content .= "PREPARE_PERCENT = ".int($test_obj->{PREPARE_PERCENT})."\n" if defined $test_obj->{PREPARE_PERCENT};
		$content .= "CURRENT_PERCENT = ".$test_obj->{PERCENT}."\n" if defined $test_obj->{PERCENT};
		$content .= "ESTIMATE_DURATION = ".$test_obj->{ESTIM_TIME}."\n" if defined $test_obj->{ESTIM_TIME};
	}
	
	# write the file using flock
	return write_string_as_file( '!'.$globals->{'status_file'}, $content );
}

#-----------------------------------------------------------------------

sub Cleanup_all {
	my ($ret_code) = @_;
	my $ok = ($ret_code == 0);
	
	inform "Cleanup...";
	
	foreach my $test_obj ( values %tests ) {
		next if !$test_obj->{STARTED};
		
		debug_inform "...cleanup ".$test_obj->name;
		
		# Check the subshell is closed
		$test_obj->Check_subshell_off();
		
		if ( $test_obj->{STATE} eq 'Running' || $test_obj->{STATE} eq 'Preparing' ) {
			$test_obj->state('Crashed');
		}
		
		$test_obj->Cleanup();
	}
	
	if ( $ok && !DEBUG ) {
		cmd( "rm -rf ".shq($globals->{'tmp_dir'}) );
	}
	
	if ( DEBUG && @all_warnings ) {
		inform "# All warnings of this run (for DEBUG purposes):";
		foreach my $warning ( @all_warnings ) {
			myprint "!!! Warning:\n".$warning->tostring()."--------------------\n";
		}
	}
	
	if ( !$ok ) {
		my $readme_file = dirname($globals->{'script_directory'})."/README";
		my $verbose_log_file = $globals->{'result_dir'}."/verbose_log";
		
		inform <<HEREDOC
#--------------------------------------------------------------
# There were errors.
# You could see the verbose log in the file 
# '$verbose_log_file'.
# The most common problems are described in the Troubleshooting
# section of the README file
# '$readme_file'
# If you believe that there is a bug in the DTK Manager,
#  please report it to [linux\@ispras.ru].
#--------------------------------------------------------------
HEREDOC
	}
	else {
		inform "Finished OK.";
	}
}
#----------------------------------------------------------------------

sub Use_profile_general {
	my ($profile, $options) = @_;
	
	# GENERAL
	my $general = $profile->{'GENERAL'};
	if ( defined $general ) {
		
		# "ARCHITECTURE" => "ia32"
		if ( defined $general->{'ARCHITECTURE'} ) {
			$host_info->{'architecture'} = guess_architecture( $general->{'ARCHITECTURE'} );
		}
	
		# "AUTOMATIC_TESTS" => "1"
		# unused
		
		# "CERTIFY" => "1"
		$options->{'cert'} = $general->{'CERTIFY'} if defined $general->{'CERTIFY'};
		
		# "STD_PROFILE" => "core,c++,desktop"
		$options->{'std-profile'} = $general->{'STD_PROFILE'} if defined $general->{'STD_PROFILE'};
		
		# STANDARD
		$options->{'standard'} = $general->{'STD_VERSION'} if defined $general->{'STD_VERSION'};
		
		# MANUAL_TESTS
		# not used here
		
		# SHORT_SCENARIOS
		# not used yet
		
		# USE_INTERNET
		if ( defined $general->{'USE_INTERNET'} && $general->{'USE_INTERNET'} ) {
			$options->{'download'} = 1;
		}
		
		# VERBOSE_LEVEL
		$options->{'verbose'} = $general->{'VERBOSE_LEVEL'} if defined $general->{'VERBOSE_LEVEL'};
		
		# NAME, ORGANIZATION
		# TODO: command line parameter
		$globals->{'tester_name'} = $general->{'NAME'} if $general->{'NAME'};
		$globals->{'tester_org'} = $general->{'ORGANIZATION'} if $general->{'ORGANIZATION'};
		# COMMENTS
		$globals->{'run_comments'} = $general->{'COMMENTS'} if $general->{'COMMENTS'};
		
		if ( $general->{'EMAIL'} && $general->{'SEND_EMAIL'} ) {
			$options->{'mail-to'} = $general->{'EMAIL'};
		}
		
	} # End of {GENERAL}
	
	# TEST_SUITES: add tests to the @tests_to_run list
	if ( defined $profile->{TEST_SUITES_SORT} ) {
		foreach my $testsuite ( @{$profile->{TEST_SUITES_SORT}} ) {
			next if !$profile->{TEST_SUITES}{$testsuite}{RUN}; # Not selected
			
			push @{$options->{TESTS}}, $testsuite;
		}
	}

	return 1;
}
#----------------------------------------------------------------------

sub list_available_versions {
	
	my $manifest = $globals->{'manifest'} or return error "Manifest isn't loaded.";
	my $architecture = $host_info->{'architecture'}; # TODO: for all back-compatible architectures
	my $standard = $globals->{'standard'};
	
	# Get information about installed packages
	is_ok packages_info( $manifest->{PACKAGES} ) or warning $Error::Last;
	check_files( values %{$manifest->{FILES}} );
	
	print "\nTests available for $standard, $architecture:\n\n";
	
	foreach my $test_name ( grep {!$manifest->{MANUAL}{$_}} @{$manifest->{ALL}{TESTSUITES}} ) {
		
		print "-------------------------------------------\n";
		
		# ->{DESCRIPTIONS}{$testsuite}[]->{STANDARD|DISPLAYNAME|DESCRIPTION} = ".."
		my ($descr) = grep {version_filter($_->{STANDARD}, $standard)} @{$manifest->{'DESCRIPTIONS'}{$test_name}};
		if ( $descr ) {
			print $test_name.": ".$descr->{DISPLAYNAME}."\n";
			print $descr->{DESCRIPTION}."\n" if $descr->{DESCRIPTION};
		} else {
			print "$test_name: this test isn't described in the Manifest.\n";
		}

		# ->{PROFILES}{$standard}{$testsuite}{$profile} = 1
		my $prof_hash = $manifest->{PROFILES}{$standard}{$test_name};
		if ( $prof_hash ) {
			print "(Profiles: ".join(' | ', sort grep { $prof_hash->{$_} } keys %$prof_hash).")\n";
		}
	
		# ->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}{$test_ver}[]->{PACKAGE|PACKAGE_VER|LOCATION} = ".."
		my $composition = $manifest->{COMPOSITION}{$standard}{$test_name}{$architecture};
		if ( !$composition ) {
			print "\n";
			print "Not available for $standard\n";
			next;
		}
		
		foreach my $status ( sort keys %$composition ) {
			my $test_vers = $composition->{$status} or next;
			foreach my $test_ver ( sort keys %$test_vers ) {
				print "\n";
				print "  Version ".$test_ver." ($architecture, $standard, $status):\n";
					
				my $files = select_files( $manifest, $test_name, $test_ver, $status, $standard, $architecture );
				if ( is_err($files) ) {
					warning "Failed to find files.\n", $Error::Last;
					next;
				}
				
				for my $file ( @$files ) {
					print "      ";
					print $file->{PACKAGE} || $file->{FILENAME};
					print " v. ".$file->{VERSION} if $file->{VERSION};
					
					my $installed = 0;
					if ( $file->{PACKAGE} && $file->{VERSION} ) {
						if ( $manifest->{PACKAGES}{$file->{PACKAGE}}{VERSIONS}{$file->{VERSION}} ) {
							$installed = 1;
						}
					}
					
					print " (";
					if ( $installed ) {
						print "installed";
					} else {
						if ( !$file->{DOWNLOADED} ) {
							print "download ".$file->{SIZE}." bytes";
						}
						elsif ( $file->{DOWNLOADED} == -1 ) {
							print "old version; new: ".$file->{SIZE}." bytes";
						}
						elsif ( $file->{DOWNLOADED} == 1 ) {
							print "downloaded";
						}
					}
					print ")";
					print "\n";
				}
			}
		}
	} continue {
		print "\n";
	}
}
#-----------------------------------------------------------------------

# Restore global variables using information from special files
#  in a specified result directory.
# This function is used when re-building report.
sub restore_env {
	my ($result_dir) = @_;
	
	$globals->{'result_dir'} = $result_dir;
	$globals->{'testrun_id'} = extract_filename($result_dir);
	$globals->{'test_result_dir'} = $result_dir."/results";
	
	# Restore runconfig
	my $runconfig = read_config( $result_dir."/runconfig" );
	is_ok($runconfig) or return error "Failed to read runconfig", $Error::Last;
	
	$globals->{'standard'} = $runconfig->{STANDARD} if $runconfig->{STANDARD};
	$globals->{'cert_mode'} = $runconfig->{CERT} if $runconfig->{CERT};
	$globals->{'run_comments'} = $runconfig->{RUN_COMMENTS} if $runconfig->{RUN_COMMENTS};
	
	if ( $runconfig->{SECTIONS} ) {
		foreach my $test_name ( keys %{$runconfig->{SECTIONS}} ) {
			my $data = $runconfig->{SECTIONS}{$test_name};
			
			push @tests_to_run, $test_name if !in_array($test_name, @tests_to_run);
			
			my $test_obj = $tests{$test_name} ||= {NAME => $test_name};
			for my $key (keys %$data) {
				$test_obj->{$key} = $data->{$key};
			}
		}
	}
	
	# Restore .info data
	my @info_files = glob($result_dir."/results/*.info");
	foreach my $info_file ( @info_files ) {
		my $test_info = read_config( $info_file );
		my $test_name = $test_info->{NAME} or next;
		
		push @tests_to_run, $test_name if !in_array($test_name, @tests_to_run);
		
		my $test_obj = $tests{$test_name} ||= {NAME => $test_name};
		for my $key (keys %$test_info) {
			$test_obj->{$key} = $test_info->{$key};
		}
	}
}

sub rebuild_report {
	my ($dir, $keep) = @_;
	
	# Restore saved environment
	is_ok restore_env($dir) or fail $Error::Last;
	
	# Load Manifest
	debug_inform "Loading manifest...";
	my $manifest = $globals->{'manifest'} = load_manifest();
	is_ok $manifest or fail "Failed to load manifest.", $Error::Last;
	
	# Load problem database
	$Report::result_dir = $globals->{'result_dir'};
	$Report::cert_mode = $globals->{'cert_mode'};
	inform "Loading problem database...";
	if ( $globals->{'refresh_prdb'} ) {
		is_ok refresh_prdb() or complain $Error::Last;
	}
	is_ok Report::load_problem_database( PROBLEM_DB_FILE )
		or warning "Failed to load the problem database.", $Error::Last;
	
	( defined $globals->{'standard'} ) or fail "Wrong data stored";
	
	# Generate test reports
	foreach my $test_name ( @tests_to_run ) {
		my $test_obj = $tests{$test_name} or next;
		if ( $keep && $test_obj->{VERDICT} && !$test_obj->{MANUAL} ) {
			debug_inform "skip $test_name: ".Report::verdict_text($test_obj->{VERDICT});
			next; # Don't rebuilt existing reports if '--keep' key was specified (except for manual tests)
		}
		
		# Load test modules and create test objects
		if ( !$test_obj->{VERSION} ) {
			warning("No VERSION stored for $test_name");
			$tests{$test_name} = undef;
			next;
		}
		# Load test module and create test object
		my $module_file = prepare_modules( $test_obj );
		if ( !is_ok($module_file) ) {
			warning "Failed to prepare modules for '$test_name'.", $Error::Last;
			$tests{$test_name} = undef;
			next;
		}
		my $class_name = load_test_module($module_file);
		if ( !is_ok($class_name) ) {
			warning "Failed to load module '$module_file'.", $Error::Last;
			$tests{$test_name} = undef;
			next;
		}
		$test_obj = $class_name->new( %$test_obj ); # create test object
		
		$test_obj->Report();
		inform $test_obj->name.": ".Report::verdict_text($test_obj->{VERDICT}) if $test_obj->{VERDICT};
	}
	
	is_ok CompileReport() or fail $Error::Last;
	
	# Make a tarball with the results
	$globals->{'results_tarball'} = $globals->{'result_dir'}."/".$globals->{'testrun_id'}.".tgz";
		
	cmd( "rm -f ".shq($globals->{'results_tarball'}) ); # Remove the old tarball
	
	my $tmp_file = extract_filename($globals->{'results_tarball'});
	cmd( "cd ".shq(extract_dir($globals->{'result_dir'}))
		." && tar czf ".shq($tmp_file)." ".shq( extract_filename($globals->{'result_dir'}) )
		." && mv ".shq($tmp_file)." ".shq($globals->{'results_tarball'})
		);

	inform "Report: ".$globals->{'result_dir'};
	inform "Verdict: ".Report::verdict_text($globals->{'verdict'}); # TODO: manual tests verdict?
}

sub CompileReport {
	inform "Making report...";

	my ($verdict, $verdict_man) = Report::Compile();
	if ( !is_ok($verdict) ) {
		complain "Failed to compile the report", $Error::Last;
		$verdict = 'run_problems';
	}
	
	if ( $verdict !~ /^(passed|warning|incomplete)$/ ) {
		$verdict = 'failed';
	}
	if ( $verdict_man !~ /^(incomplete|passed)$/ ) {
		$verdict_man = 'failed';
	}
	
	$globals->{'verdict'} = $verdict;
	$globals->{'verdict_man'} = $verdict_man;
	
	write_INFO();
}
#-----------------------------------------------------------------------

sub send_results {
	my ( $tarball_file ) = @_;
	
	( $globals->{'mailto'} ) or return error "Mailto parameter isn't defined.";
	( $tarball_file ) or return error "Results tarball missed";
	
	# Send the results by e-mail if requested.
	inform "Sending the results by e-mail...\n";

	# Use sendmail utility
	sendmail (
		TO      => $globals->{'mailto'},
		SUBJECT => "LSB Distribution Testkit results for ".$host_info->{'hostname'},
		TEXT    => "Please see the attached archive for the test results.",
		ATTACHMENT => $tarball_file,
	);
}
#-----------------------------------------------------------------------

sub refresh_prdb {
    return download_file(PROBLEM_DB_LINK, PROBLEM_DB_FILE);
}

#============================================================================
# The Main Script
#============================================================================

# Initialization
is_ok Setup()
	or fail "Initialization failed.", $Error::Last;

# Write the status file.
unless ( $globals->{'check_only'} ) {
	is_ok write_status()
		or fail "Can't write status file", $Error::Last;
	
	print STDERR "Started.\n" if $globals->{'webui'};
	# Since this point the WEB-UI shows the console output.
}
	
$_END = sub {
	if ( $need_cleanup ) {
		Cleanup_all( $? );
	}
	
	$globals->{'done'} ||= 'Terminated';
	write_status();
	
	print STDERR "Finished.\n" if $globals->{'webui'};
};

# Run test checks
{
	foreach my $test_name ( @tests_to_run ) {
		my $test_obj = $tests{$test_name};

		inform "Checking $test_name...";
		
		my $res = $test_obj->check();
		( is_ok($res) && $res ) or fail "The $test_name check failed.", $Error::Last;
	}
}

if ( $globals->{'check_only'} ) {
	inform "Done.";
	exit 0;
}

# Setup interrupt handlers
$SIG{INT} = $SIG{TERM} = sub { 
	$SIG{TERM} = 'IGNORE'; 
	
	fail "Caught a terminate signal!";
};

write_status();

# TODO: Stop the ntpd???

unless ( DEBUG && @tests_to_run == 1 ) { # DBG:
	inform "Benchmarking...";
	Benchmark();
	inform "...done.";
}

# Prepare the report module
{
	if ( $globals->{'refresh_prdb'} ) {
		is_ok refresh_prdb() or complain $Error::Last;
	}
	$Report::result_dir = $globals->{'result_dir'};
	$Report::cert_mode = $globals->{'cert_mode'};
	inform "Loading problem database...";
	is_ok Report::load_problem_database( PROBLEM_DB_FILE )
		or warning "Failed to load the problem database."
			." Please update the problem database using option '--refresh-prdb'.", $Error::Last;
	inform "...done.";
}

# Write the run configuration  [after the Benchmark]
{
	write_status();
	# Save the system configuration
	is_ok write_runconfig()
		or fail "Failed to save configuration.", $Error::Last;
	# Write the info file
	is_ok write_INFO()
		or fail "Failed to write INFO file.", $Error::Last;
	
	# Save the test configuration
	foreach my $test_name ( @tests_to_run ) {
		my $test_obj = $tests{$test_name};
		is_ok $test_obj->save_info()
			or fail "Failed to save configuration for test '$test_name'", $Error::Last;
	}
}

## RUN THE TESTS ##

# Stop the running test if captured the Interrupt signal.
$SIG{INT} = sub {
	if ( $globals->{'interrupt_request'} ) { # Ctrl-C pushed twice
		kill 'TERM', $$; # send the TERM signal to itself.
		return;
	}
	$globals->{'interrupt_request'} = 1;
	
	inform "Caught an interrupt signal!";
	
	my $subshell_stopped = 0;
	foreach my $test_obj ( values %tests ) {
		next if !$test_obj->{STARTED};
		
		# Force the subshell to be closed
		if ( $test_obj->Check_subshell_off() ) {
			$subshell_stopped = 1;
		}
	}
	if ( !$subshell_stopped ) {
		kill 'TERM', $$; # send the TERM signal to itself.
		return;
	}
};

write_status();

$need_cleanup = 1; # Used in END{}

my $test_obj = undef; # (Used in the "continue" section)

# THE MAIN CYCLE
foreach my $test_name ( @tests_to_run ) {
	$test_obj = $tests{$test_name};
	
	inform "==========================================";
	
	$globals->{'interrupt_request'} = 0;
	
	is_ok $test_obj->Go()
		or complain $Error::Last;
	
	$globals->{'interrupt_request'} = 0; # TODO: use in no-subshell app-bat tests
	
	$test_obj->Report();
	# Print the test verdict
	inform $test_obj->name.": ".Report::verdict_text($test_obj->{VERDICT}) if $test_obj->{VERDICT};
	
	write_runconfig();
	
	# Compile the report
	CompileReport();
}
#-----------------

# Finish
inform "==========================================";

# Make a tarball with the results
{
	$globals->{'results_tarball'} = $globals->{'result_dir'}."/".$globals->{'testrun_id'}.".tgz";
	
	my $tmp_file = $globals->{'temp_dir'}."/".extract_filename($globals->{'results_tarball'});
	cmd( "cd ".shq(extract_dir($globals->{'result_dir'}))
		." && tar czf ".shq($tmp_file)." ".shq( extract_filename($globals->{'result_dir'}) )
		." && mv ".shq($tmp_file)." ".shq($globals->{'results_tarball'})
		);
}

inform "Report: ".$globals->{'result_dir'};
inform "Verdict: ".Report::verdict_text($globals->{'verdict'}); # TODO: manual tests verdict?

if ( $globals->{'mailto'} ) {
	# Send results by e-mail
	is_ok send_results( $globals->{'results_tarball'} )
		or complain $Error::Last;
}

$globals->{'done'} = 'Finished';

exit 0;

END { goto &$_END if $_END; }
