#!/usr/bin/perl -w

# LSB Test Execution Framework
# Manifest Generator (manifest_gen.pl)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use strict;
use FindBin;

sub BEGIN {
	# Add current directory to the @INC to be able to include modules
	# from the same directory where this script is located.
	unshift @INC, $FindBin::Bin;
	
	$| = 1; # Immediate flush of all output
}

use Common;
use Manifest;
use Ftp;

use Download;
use Packages;

#----------------------------------------------------------------------

my @errors;

#----------------------------------------------------------------------

my $check_unique_name = 0;

#----------------------------------------------------------------------

my $directory = CACHE_DIR; # Scan the directory

my $ftp = undef; # Scan the FTP

my $output_filename = undef;

my $DEFAULT_MANIFEST_URL = 'http://ftp.linuxfoundation.org/pub/lsb/updates/dtk-manager-data/Manifest';

#----------------------------------------------------------------------

sub make_file {
	my ($filename, $size, $timestamp) = @_;
	
	( defined $filename )
		or return error "make_file: undefined parameter";
	
	my ($package_name, $package_ver);
	my $arch = '*';
	my $kind = 'file';
	
	# Parse the file name
	if ( $filename =~ m/\.(rpm|deb)$/ ) {
		if ( $filename =~ m/^([\w_\+\-]+)-([\w\.\-]+)\.(\w+)\.rpm$/ ) {
			$package_name = $1;
			$package_ver = $2;
			return if $3 eq 'src';
			$arch = guess_architecture($3, "*");
			$kind = 'rpm';
		}
		elsif ( $filename =~ m/^([\w\+\-]+)_([\w\.\-]+)_(\w+)\.deb$/ ) {
			$package_name = $1;
			$package_ver = $2;
			return if $3 eq 'src';
			$arch = guess_architecture($3, "*");
			$kind = 'deb';
		}
		else {
			return error "Can't parse file name: '$filename'";
		}
	}
	elsif ( $filename =~ m/\.(tar|pdf)$/ ) { # .tar and .pdf are data files for some tests.
		$package_name = $filename;
		$package_ver = "";
		$kind = 'file';
		$arch = '*';
	}
	else {
		return if $filename =~ /\.(txt)$/i; # skip warnings for txt files
		return if $filename =~ /^[^\.]*$/i; # skip files without "extension"
		return if $filename =~ /^Licence\./; # skip "Licence." files
		
		warning "Skipped: '$filename'";
		return undef;
	}
	
	( defined $arch ) or return error "Undefined arch: '$filename'";
	
	# Hacks
	if ( $package_name && $package_name eq "lsb-tet3" ) {
		$package_name = "lsb-tet3-lite";
		$package_ver =~ s/\.lsb3$//;
	}
	
	my $res = {};
	
	$res->{FILENAME} = $filename;
	$res->{SIZE}     = $size;
	$res->{TIMESTAMP} = $timestamp;

	$res->{ARCH}    = $arch;

	$res->{KIND}    = $kind; # rpm|deb|file
	$res->{PACKAGE} = $package_name;
	$res->{VERSION} = $package_ver;
	
	return $res;
}

sub print_file_line {
	my ($file) = @_;
	
	my $res = "";
	
	$res .= $file->{FILENAME};
	$res .= "\t".$file->{LOCATION};
	$res .= "\t"."rm" if $file->{REMOVED};
	$res .= "\t".$file->{SIZE};
	$res .= "\t"."t".$file->{TIMESTAMP} if defined $file->{TIMESTAMP};
	$res .= "\t".$file->{ARCH};
	
	$res .= "\t".$file->{KIND};
	$res .= "\t".$file->{PACKAGE};
	if ( $file->{KIND} ne 'file' ) {
		$res .= "\t".$file->{VERSION} if $file->{VERSION};
	}
	
	return $res;
}
#----------------------------------------------------------------------
sub read_dir {
	my ($dir) = @_;
	
	( -d $dir ) or return error "'$dir' doesn't exist or not a directory";
	
	opendir DIR, $dir  or return error "Can't open directory '$dir': $!";
	my @dir_list = readdir(DIR) or return error "Can't read directory '$dir': $!";
	
	my @files = ();
	
	foreach my $filename ( @dir_list ) {
		
		next if $filename =~ /^\./;  # Skip hidden files, "." and ".."
		next if $filename !~ m/\.(rpm|deb|tar|pdf)$/; # Select only interesting files.
		
		my $fullname = $dir."/".$filename;
		
		next if !-f $fullname;   # Skip everything that isn't a file or a symlink
		
		my ($size, $mtime) = file_info($fullname);
		( defined $size ) or next;
		
		my $file = make_file( $filename, $size, $mtime );
		if ( is_err($file) ) {
			warning "Failed to add file: $filename, $size, $mtime", $Error::Last;
			next; # Skipped
		}
		
		push @files, $file;
	}
	return @files;
}

sub read_ftp {
	my ($srv, $dir, $cache_dir) = @_;
	
	my @ftp_list = Ftp::list_ftp ( $srv, $dir, $cache_dir );
	
	my @files = ();
	
	foreach ( @ftp_list ) {
		my $file = make_file( @$_ );
		if ( is_err($file) ) {
			warning "Failed to add file: @$_", $Error::Last;
			next; # Skipped
		}
		
		push @files, $file;
	}
	
	return @files;
}
#----------------------------------------------------------------------

sub check_manifest {
	my ($manifest, %options) = @_;
	
	@errors = ();
	
	# Check MANIFEST section
	if ( !$manifest->{MANIFEST}{VERSION} ) {
		return error "No VERSION in the MANIFEST section";
	}
	if ( $manifest->{MANIFEST}{VERSION} ne MANIFEST_FORMAT_VER ) {
		return error "Unsupported manifest version '".$manifest->{MANIFEST}{VERSION}."'"
				." The only supported version is ".MANIFEST_FORMAT_VER;
	}
	
	if ( !$manifest->{MANIFEST}{URL} ) {
		return error "The MANIFEST section should specify the URL.";
	}
	
	# Check DESCRIPTION
	foreach my $standard ( keys %{$manifest->{TREE}} ) {
	foreach my $testsuite ( keys %{$manifest->{TREE}{$standard}} ) {
		# ->{DESCRIPTIONS}{$testsuite}[]->{STANDARD|DISPLAYNAME|DESCRIPTION} = ".."
		my @descr = grep {version_filter($_->{STANDARD}, $standard)} @{$manifest->{'DESCRIPTIONS'}{$testsuite}};
		if ( !@descr ) {
			push @errors, error "No DESCRIPTION for '$testsuite': '$standard'";
		}
		elsif ( @descr > 1 ) {
			if ( $options{'strict'} ) {
				push @errors, error "Multiple DESCRIPTIONS for '$testsuite': '$standard' (".join(", ", map {$_->{STANDARD}} @descr).")";
			}
		}
		else {
			my ($descr) = @descr;
			( $descr->{DISPLAYNAME} )
				or push @errors, error "No DISPLAYNAME for '$testsuite': '$standard'";
			( $descr->{DESCRIPTION} )
				or push @errors, error "No DESCRIPTION for '$testsuite': '$standard'";
		}
	}}
	# ->{DESCRIPTIONS}{$testsuite}[]->{STANDARD|DISPLAYNAME|DESCRIPTION} = ".."
	foreach my $testsuite ( keys %{$manifest->{DESCRIPTIONS}} ) {
		( in_array($testsuite, @{$manifest->{ALL}{TESTSUITES}}) )
			or push @errors, error "Unknown '$testsuite' at DESCRIPTIONS.";
	}
	
	# Check PROFILES
	foreach my $standard ( keys %{$manifest->{PROFILES}} ) {
	foreach my $testsuite ( keys %{$manifest->{PROFILES}{$standard}} ) {
		# ->{PROFILES}{$standard}{$testsuite}{$profile_id} = 1
		foreach my $profile_name ( keys %{$manifest->{PROFILES}{$standard}{$testsuite}} ) {
			( defined $manifest->{STD_PROFILES}{$standard}{$profile_name} )
				or push @errors, error "Profile '$profile_name' specified for '$testsuite' is not allowed for '$standard'. See the beginning of the PROFILES section."
						." Allowed profiles for '$standard' are: ".join(";", map "'$_'", keys %{$manifest->{STD_PROFILES}{$standard}});
		}
	}}
	
	# CHECK [COMPOSITION]
	#  ->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}{$test_ver}[]->{PACKAGE|PACKAGE_VER|LOCATION} = ".."
	my $composition = $manifest->{COMPOSITION};
	foreach my $standard  ( keys %{$composition} ) {
	foreach my $testsuite ( keys %{$composition->{$standard}} ) {
	foreach my $arch      ( keys %{$composition->{$standard}{$testsuite}} ) {
		my $status_list = $composition->{$standard}{$testsuite}{$arch};
	foreach my $status    ( keys %$status_list ) {
	foreach my $test_ver  ( keys %{$status_list->{$status}} ) {
		my $reqs = $status_list->{$status}{$test_ver};
		next if $test_ver eq '*'; # Unavailable test
		foreach my $req ( @$reqs ) {
			my $package_name = $req->{PACKAGE};
			my $package_ver  = ( $req->{PACKAGE_VER} or "*" );
			my $location     = ( $req->{LOCATION} or "" );
			
			my $found = 0;
			foreach my $package_manager ( "rpm", "dpkg" ) {

				my $found_files = Manifest::select_file( $manifest, $arch, $package_manager,
							$package_name, $package_ver, $location);
				if ( is_err($found_files) || !$found_files || !@$found_files ) {
					# It's no good if there is no rpm package.
					if ( $package_manager eq 'rpm' && $options{'strict'} ) {
						push @errors, error "check: Can't find RPM package '$package_name' v. '$package_ver'"
								.", '$standard', arch: '$arch', required by '$testsuite'"
								.($location ? " in '$location'": ""), $Error::Last;
					}
					next;
				}
				if ( @$found_files > 1 && $options{'strict'} ) {
					push @errors, error  "Ambigous file choise: ".$found_files->[0]->{SRC}
						.": ".join ", ", map {$_->{FILENAME}} @$found_files;
				}
				
				$found = 1;
			} # end of  foreach my $package_manager ( "rpm", "dpkg" )
			if ( !$found ) { 
				push @errors, error "check: Can't find file '$package_name' v. '$package_ver'"
						.", '$standard', arch: '$arch', required by '$testsuite'"
						.($location ? " in '$location'": ""), $Error::Last;
			}
		} # end of  foreach my $req ( @$reqs )
	}}}}} # end of  foreach {$testsuite}{$standard}{$status}{$arch}{$test_ver}
	
	# CHECK [MODULES]
	{
		my $versions = {};
		foreach my $standard  ( keys %$composition ) {
		foreach my $testsuite ( keys %{$composition->{$standard}} ) {
		foreach my $arch      ( keys %{$composition->{$standard}{$testsuite}} ) {
			foreach my $status    ( keys %{$composition->{$standard}{$testsuite}{$arch}} ) {
				foreach my $test_ver  ( keys %{$composition->{$standard}{$testsuite}{$arch}{$status}} ) {
					$versions->{$testsuite}{$standard}{$test_ver} = 1;
				}
			}
		}}}
		foreach my $testsuite ( keys %$versions ) {
			if ( !defined $manifest->{MODULES}{$testsuite} ) {
				push @errors, error "No MODULES for '$testsuite'";
				next;
			}
			foreach my $standard  ( keys %{$versions->{$testsuite}} ) {
			foreach my $test_ver  ( keys %{$versions->{$testsuite}{$standard}} ) {
				my @modules = grep {
						   version_filter($_->{TEST_VER}, $test_ver)
						&& version_filter($_->{STANDARD}, $standard)
					} @{$manifest->{MODULES}{$testsuite}};
				if ( !@modules ) {
					push @errors, error "No MODULES for '$testsuite' ver $test_ver, '$standard'";
					next;
				}
				@modules = grep $_->{MAIN}, @modules;
				if ( scalar @modules != 1 ) {
					push @errors, error "Should be one main MODULE for '$testsuite' ver $test_ver, '$standard'";
					next;
				}
			}}
		}
	}
	
	# Check OPTIONS
	# ->{OPTIONS}{$testsuite}[]->{ID|Q|GLOBAL=>1|TEST_VER|STANDARD|PRE|TYPE|VARIANTS->[]|DEFAULT|VALUE};
	if ( defined $manifest->{OPTIONS} ) {
		my $used_options = {};
		foreach my $testsuite ( keys %{$manifest->{OPTIONS}} ) {
			foreach my $opt ( @{$manifest->{OPTIONS}{$testsuite}} ) {
				( defined $opt->{ID} )
					or push @errors, error "Option without ID: '$testsuite'."; # Should not happen
				my $id = $opt->{ID};
				( defined $opt->{STANDARD} && $opt->{STANDARD} =~ /\S/ )
					or push @errors, error "Standard isn't specified for OPTION: '$testsuite' '$id'.";
				if ( $opt->{GLOBAL} && !$opt->{DEFAULT} ) {
					push @errors, error "Global option without default value: ID '$id' for '$testsuite'.";
				}
				if ( $opt->{GLOBAL} && $opt->{TYPE} !~ /^(enum|bool)$/ ) {
					push @errors, error "Global option of unsupported type (only 'enum' and 'bool' types are allowed by now): ID '$id' for '$testsuite'.";
				}
				if ( !$opt->{GLOBAL} && !$opt->{Q} ) {
					push @errors, error "Option without the Q value (short text description): ID '$id' for '$testsuite'.";
				}
				my $key = $id."|";
				$key .= "VER ".$opt->{TEST_VER} if $opt->{TEST_VER};
				$key .= "|".($opt->{STANDARD} or "");
				$used_options->{$testsuite}{$key} = 0;
			}
		}
		my $versions = {};
		foreach my $standard  ( keys %$composition ) {
		foreach my $testsuite ( keys %{$composition->{$standard}} ) {
		foreach my $arch      ( keys %{$composition->{$standard}{$testsuite}} ) {
			next if !$manifest->{OPTIONS}{$testsuite};
			
			foreach my $status    ( keys %{$composition->{$standard}{$testsuite}{$arch}} ) {
				foreach my $test_ver  ( keys %{$composition->{$standard}{$testsuite}{$arch}{$status}} ) {
					$versions->{$testsuite}{$standard}{$test_ver} = 1;
				}
			}
		}}}
		foreach my $testsuite ( keys %$versions ) {
		foreach my $standard  ( keys %{$versions->{$testsuite}} ) {
		foreach my $test_ver  ( keys %{$versions->{$testsuite}{$standard}} ) {
			my $test_opts = {};
			for my $opt ( grep { version_filter($_->{TEST_VER}, $test_ver) }
						  grep { version_filter($_->{STANDARD}, $standard) }
						  @{$manifest->{OPTIONS}{$testsuite}} )
			{
				my $id = $opt->{ID};
				if ( $test_opts->{$id} ) {
					my $v1 = ($opt->{TEST_VER} ? "VER ".$opt->{TEST_VER} : "")."|".($opt->{STANDARD} or "");
					my $v2 = $test_opts->{$id};
					if ( $options{'strict'} ) {
						push @errors, error "Double option for '$testsuite': '$v1' == '$v2' for VER '$test_ver' '$standard', ID '$id'.";
					}
				}
				$test_opts->{$id} = ($opt->{TEST_VER} ? "VER ".$opt->{TEST_VER} : "")."|".($opt->{STANDARD} or "");
				my $key = $id."|";
				$key .= "VER ".$opt->{TEST_VER} if $opt->{TEST_VER};
				$key .= "|".($opt->{STANDARD} or "");
				$used_options->{$testsuite}{$key} = 1;
			}
		}}}
		if ( $options{'strict'} ) {
			foreach my $testsuite ( keys %{$manifest->{OPTIONS}} ) {
				my @unused = grep {$used_options->{$testsuite}{$_}==0} keys %{$used_options->{$testsuite}};
				if ( @unused ) {
					warning "Unused options for '$testsuite': ".join(", ", @unused);
				}
			}
		}
	} # end of  if ( defined $manifest->{OPTIONS} ) 
	
	# Check MANUAL
	# ->{MANUAL}{$testsuite}[]->{STEP|STANDARD}
	if ( defined $manifest->{MANUAL} ) {
		my $used_manual = {};
		foreach my $testsuite ( keys %{$manifest->{MANUAL}} ) {
			foreach my $step ( @{$manifest->{MANUAL}{$testsuite}} ) {
				( defined $step->{STEP} )
					or push @errors, error "Manual step without ID: '$testsuite'."; # Should not happen
				my $text = $step->{STEP};
				my $text_short = substr($text, 0, 80); # for error messages
				( defined $step->{STANDARD} && $step->{STANDARD} =~ /\S/ )
					or push @errors, error "Standard isn't specified for MANUAL test step: '$testsuite' '$text_short'.";
				my $key = $text."|".($step->{STANDARD} or "");
				$used_manual->{$testsuite}{$key} = 0;
			}
		}
		my $versions = {};
		foreach my $standard  ( keys %$composition ) {
		foreach my $testsuite ( keys %{$composition->{$standard}} ) {
		foreach my $arch      ( keys %{$composition->{$standard}{$testsuite}} ) {
			next if !$manifest->{MANUAL}{$testsuite};
			
			$versions->{$testsuite}{$standard} = 1;
		}}}
		foreach my $testsuite ( keys %$versions ) {
		foreach my $standard  ( keys %{$versions->{$testsuite}} ) {
			my $test_steps = {};
			for my $step ( grep { version_filter($_->{STANDARD}, $standard) }
						  @{$manifest->{MANUAL}{$testsuite}} )
			{
				my $text = $step->{STEP};
				my $text_short = substr($text, 0, 80); # for error messages
				if ( $test_steps->{$text} ) {
					my $v1 = $step->{STANDARD} || "";
					my $v2 = $test_steps->{$text};
					if ( $options{'strict'} ) {
						push @errors, error "Double manual step for '$testsuite': '$v1' == '$v2' for '$standard', STEP '$text_short'.";
					}
				}
				$test_steps->{$text} = $step->{STANDARD} || "";
				my $key = $text."|".($step->{STANDARD} or "");
				$used_manual->{$testsuite}{$key} = 1;
			}
		}}
		if ( $options{'strict'} ) {
			foreach my $testsuite ( keys %{$manifest->{MANUAL}} ) {
				my @unused = grep {$used_manual->{$testsuite}{$_}==0} keys %{$used_manual->{$testsuite}};
				for my $s ( @unused ) {
					my $standard = "";
					if ( $s =~ s/\|([^\|]*)$// ) { $standard = $1; }
					warning "Unused manual step for '$testsuite': $standard, '$s'";
				}
			}
		}
	} # end of  if ( defined $manifest->{MANUAL} ) 
	
	if ( @errors ) { return join_errors(@errors); }
	
	return 1;
}
#----------------------------------------------------------------------

sub list_files {
	my ($manifest) = @_;
	
	my %locations = (); # $locations{"/dir/"} = "<location>"
	
	foreach my $location ( sort keys %{$manifest->{SCAN}} ) {
		my $dir = expand_aliases( $location, $manifest->{LOCATIONS} );
		is_ok($dir) or return error "Failed to expand aliases in '$location'", $Error::Last;
		if ( $dir =~ /[<>]/ ) {
			foreach my $arch ( @all_archs ) {
				my $dir2 = expand_aliases( $dir, $manifest->{LOCATIONS}, $arch );
				is_ok($dir2) or return error "Failed to expand aliases in '$dir'", $Error::Last;
				( $dir2 !~ /[<>]/ ) or return error "Failed to expand aliases in '$dir': $dir2", $Error::Last;
				$locations{$dir2} = $location;
			}
		}
		else {
			$locations{$dir} = $location;
		}
	}
	
	if ( $directory ) {
		foreach my $src ( sort keys %locations ) {
			my $location = $locations{$src};
			
			my $dir = $src;
			$dir =~ s{^/[^/]*}{}; # strip the beginning
			$dir = $directory.$dir;
			print "> $dir\n";
			my @files = read_dir($dir);
			is_ok(@files) or warning("Failed to read dir '$dir'", $Error::Last), next;
			
			foreach (@files) {
				( defined ) or next;
				$_->{LOCATION} = $location;
				
				$_->{SRC} = $src;
				
				#debug_inform "    + ".$_->{FILENAME}." (".$_->{KIND}." ".$_->{PACKAGE}.")";
				
				my $key = $_->{SRC}."/".$_->{FILENAME};
				$manifest->{FILES}{$key} = $_; # Add to Manifest
			}
		}
	}
	elsif ( $ftp ) {
		foreach my $src ( sort keys %locations ) {
			my $location = $locations{$src};
			
			my $url = "ftp:/".$src;
			
			# Extract the host name and the directory at the server
			my $host;
			my $host_dir;
			if ( $url =~ /^\w+:\/\/([^\/]+)(\/.*)?/ ) {
				$host = $1;
				$host_dir = ($2 or "");
			}
			
			( $host && $host =~ /^[\w\.\-]+$/ )
				or return error "Wrong host: '$host'";
			
			$host_dir =~ s/^\///; # remove slash
			# now we got the directory name
			
			my @files = read_ftp( $host, $host_dir, cache_dest($src) ); # new files
			
			foreach (@files) {
				$_->{LOCATION} = $location;
				
				$_->{SRC} = $src;
				
				my $key = $_->{SRC}."/".$_->{FILENAME};
				$manifest->{FILES}{$key} = $_; # Add to Manifest
			}
		}
	}
	else {
		return error "--ftp or --dir should be specified";
	}
	
	# Build PACKAGES hash
	foreach my $key ( keys %{$manifest->{FILES}} ) {
		my $package = $manifest->{FILES}{$key}{PACKAGE} or next;
		push @{$manifest->{PACKAGES}{$package}{FILES}}, $key;
	}
	
	return 1;
}
#----------------------------------------------------------------------

sub main {
	my @files = @_;

	# Initialize the manifest with an empty structure
	my $manifest = {};
	
	# Manifest per test module.
	foreach my $manifest_file ( @files ) {
		( -f $manifest_file ) or return error "Can't find file '$manifest_file'";
		
		my $new_manifest = manifest_parse( $manifest_file );
		( is_ok($new_manifest) && $new_manifest ) 
			or return error "Failed to parse the manifest file '$manifest_file'.", $Error::Last;
		
		delete $new_manifest->{FILES};
		
		# Merge the manifests
		is_ok merge_hash( $manifest, $new_manifest ) 
			or return error "Conflict in '$manifest_file'", $Error::Last;
	}

	( defined $manifest->{SCAN} ) or return error "No locations was used in the COMPOSITION section";

	is_ok list_files( $manifest )
		or return error "Failed to obtain the file list.", $Error::Last;

	# Resolve "*" versions in the COMPOSITION section
	is_ok auto_composition( $manifest ) or return error "Failed to resolve the manifest template.", $Error::Last;
	
	# Remove unavailable tests.
	remove_unavailable($manifest, 'verbose');
	
	# Check the result
	is_ok check_manifest( $manifest, 'strict' => 1 )
		or return error "Manifest check failed", $Error::Last;
	
	return $manifest;
}

#----------------------------------------------------------------------

# Remove unavailable tests and unneeded records from the Manifest structure.
sub remove_unavailable {
	my ($manifest, $verbose) = @_;
	
	# Filter COMPOSITION
	foreach my $standard  ( keys %{$manifest->{COMPOSITION}} ) {
	foreach my $testsuite ( keys %{$manifest->{COMPOSITION}{$standard}} ) {
	foreach my $arch      ( keys %{$manifest->{COMPOSITION}{$standard}{$testsuite}} ) {
		my $status_list = $manifest->{COMPOSITION}{$standard}{$testsuite}{$arch};
	foreach my $status    ( keys %{$status_list} ) {
	foreach my $test_ver  ( keys %{$status_list->{$status}} ) {
	
		my $requirements = $status_list->{$status}{$test_ver};
	
		foreach my $req ( @$requirements ) {
			my $package_name = $req->{PACKAGE};
			my $package_ver  = ( $req->{PACKAGE_VER} or "*" );
			my $location     = ( $req->{LOCATION} or "" );
			
			my $found_files;
			my $found = 0;
			for my $package_manager ( 'rpm', 'deb' ) {
				$found_files = Manifest::select_file( $manifest, $arch, $package_manager,
							$package_name, $package_ver, $location );
				if ( is_ok($found_files) && $found_files && @$found_files ) {
					$found = 1;
					$found_files->[0]->{USED} = 1;
				}
			}
			if ( !$found ) {
				# This test suite is not available
				delete $manifest->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}{$test_ver};
				if ( $verbose ) {
					inform "Unavailable: $standard, $testsuite $test_ver, $arch, $status:"
						." $package_name $package_ver @ "
						.expand_aliases($location, $manifest->{LOCATIONS}, $arch);
				}
				last;
			}
		}
	}
	( keys %{$manifest->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}} > 0 )
		or delete $manifest->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status};
	}
	( keys %{$manifest->{COMPOSITION}{$standard}{$testsuite}{$arch}} > 0 )
		or delete $manifest->{COMPOSITION}{$standard}{$testsuite}{$arch};
	}
	( keys %{$manifest->{COMPOSITION}{$standard}{$testsuite}} > 0 )
		or delete $manifest->{COMPOSITION}{$standard}{$testsuite};
	}
	( keys %{$manifest->{COMPOSITION}{$standard}} > 0 )
		or delete $manifest->{COMPOSITION}{$standard};
	}
	
	# Delete unused files and check for filename conflicts
	foreach my $key ( sort keys %{$manifest->{FILES}} ) {
		my $file = $manifest->{FILES}{$key};
		if ( !$file->{USED} ) {
			delete $manifest->{FILES}{$key};
			next;
		}
	}
	# Rebuild PACKAGES hash
	$manifest->{PACKAGES} = {};
	foreach my $key ( keys %{$manifest->{FILES}} ) {
		my $package = $manifest->{FILES}{$key}{PACKAGE} or next;
		push @{$manifest->{PACKAGES}{$package}{FILES}}, $key;
	}
	
	# Filter PROFILES
	foreach my $standard ( keys %{$manifest->{PROFILES}} ) {
		( defined $manifest->{COMPOSITION}{$standard} )
			or delete($manifest->{PROFILES}{$standard}), next;
		foreach my $testsuite ( keys %{$manifest->{PROFILES}{$standard}} ) {
			( defined $manifest->{COMPOSITION}{$standard}{$testsuite} )
				or delete($manifest->{PROFILES}{$standard}{$testsuite}), next;
		}
	}
	
	# ! Keep the TREE structure as is. Unavailable tests still present in the TREE structure
	# are shown by the UI as unavailable.
	
	# Let DESCRIPTIONS, OPTIONS, MANUAL sections unfiltered.
}

#----------------------------------------------------------------------

# Resolve "automatic" versions in the COMPOSITION section
sub auto_composition {
	my ($manifest) = @_;
	
	#  ->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}{$test_ver}[]->{PACKAGE|PACKAGE_VER|LOCATION} = ".."
	
	foreach my $standard  ( keys %{$manifest->{COMPOSITION}}  ) {
	foreach my $testsuite ( keys %{$manifest->{COMPOSITION}{$standard}}  ) {
	foreach my $arch      ( keys %{$manifest->{COMPOSITION}{$standard}{$testsuite}}  ) {
	foreach my $status    ( keys %{$manifest->{COMPOSITION}{$standard}{$testsuite}{$arch}}  ) {
		
		my $test_vers = $manifest->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status};
		my $req_arr = $test_vers->{'*'} or next;
		
		my @chosen_pack_vers;
		my $skip = 0;
		foreach my $req ( @$req_arr ) {
			my $package  = $req->{PACKAGE};
			my $location = $req->{LOCATION}
				or return error "Location is not defined in a req '$package'"
						." of '$standard', '$testsuite', '$arch', '$status'";
			
			my $version;
			if ( defined $req->{PACKAGE_VER} && $req->{PACKAGE_VER} !~ /^(\*|)$/ ) {
				$version = $req->{PACKAGE_VER};
			}
			else {
				# Look up a rpm or deb package
				my $found_files = Manifest::select_file( $manifest, $arch, '', $package, '*', $location );
				if ( !is_ok($found_files) || !$found_files || !@$found_files ) {
					# File not found
					# The test is unavailable
					
					#warning "$testsuite is not available for $arch, $standard, $status:"
							#." missing $package ($location)", $Error::Last;
					
					$skip = 1;
					last;
				}
				
				@$found_files = sort {vercmp($b->{VERSION}, $a->{VERSION})} @$found_files; # sort by version DESCENDING
				
				# Choose file with max version (first in the sorted list);
				$version = $found_files->[0]->{VERSION};
			}
			
			push @chosen_pack_vers,
				{ 
					PACKAGE => $package, 
					PACKAGE_VER => $version, 
					LOCATION => $location,
				};
		}
		
		# Delete the template records
		delete $test_vers->{'*'};
		
		if ( $skip ) {
			debug_inform "Can't select test suite version"
				." for $arch, '$standard', '$testsuite', status '$status'.";
			
			next;
		}
		
		# Determine the testsuite version by version of the last package
		my $last_ver = $chosen_pack_vers[-1]->{PACKAGE_VER};
		
		$last_ver =~ s/\.lsb\d$//; # remove .lsb3 suffix of version
		
		# Create a new record
		$manifest->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}{$last_ver} = [@chosen_pack_vers];
		
	}}}} # end of  foreach $standard, $testsuite, $arch, $status
}
#----------------------------------------------------------------------

sub print_manifest {
	my ($manifest) = @_;
	
	#my $manifest_url = "http://...";
	#my $hash = ???;
	
	my $res = "";
	
	# [MANIFEST]
	$res .= "[MANIFEST]"."\n";
	$res .= "VERSION = ".MANIFEST_FORMAT_VER."\n";
	$res .= "URL = ".$manifest->{MANIFEST}{URL}."\n" if $manifest->{MANIFEST}{URL};
	$manifest->{MANIFEST}{TIMESTAMP} ||= time();
	$res .= "TIMESTAMP = ".$manifest->{MANIFEST}{TIMESTAMP}."\n";
	#$res .= "HASH = ".$hash."\n";
	$res .= "\n"; # -----
	
	# [TREE]
	$res .= "[TREE]"."\n";
	# $manifest->{TREE}{$standard}{$testsuite} = "<..>/<..>"
	{
		my @lines = ();
		foreach my $standard ( keys %{$manifest->{TREE}} ) {
		foreach my $testsuite ( keys %{$manifest->{TREE}{$standard}} ) {
			my $path = $manifest->{TREE}{$standard}{$testsuite};
			# <LSB 3.0>/<Certification Tests>/<Automated Application Battery Tests>/expect
			push @lines, "<".$standard.">/".$path."/".$testsuite;
		}}
		$res .= $_."\n" foreach ( sort @lines );
	}
	$res .= "\n"; # -----

	# [DESCRIPTIONS]
	$res .= "[DESCRIPTIONS]"."\n";
	$res .= "## Format:\n";
	$res .= "# TESTSUITE testname [Standard (<=|<|=|>=|>) ver]"."\n"; 
	$res .= "#     DISPLAYNAME = ..."."\n"; 
	$res .= "#     DESCRIPTION = description ..."."\n"; 
	$res .= "# ...can be multiline..."."\n"; 
	$res .= "# [TESTSUITE END]"."\n"; 
	# $manifest->{DESCRIPTIONS}{$testsuite}->{STANDARD|DISPLAYNAME|DESCRIPTION} = ".."
	foreach my $testsuite ( sort keys %{$manifest->{DESCRIPTIONS}} ) {
		my $descriptions;
		foreach ( @{$manifest->{DESCRIPTIONS}{$testsuite}} ) {
			my $standard_spec = ( $_->{STANDARD} or "" );
			$descriptions->{$standard_spec} = $_;
		}
		
		foreach my $standard_spec ( sort keys %$descriptions ) {
			my $obj = $descriptions->{$standard_spec};
			$res .= "TESTSUITE $testsuite"."\t".$standard_spec."\n";
			$res .= "\tDISPLAYNAME = ".$obj->{DISPLAYNAME}."\n";
			$res .= "\tDESCRIPTION = ".$obj->{DESCRIPTION}."\n";
			$res .= "TESTSUITE END"."\n";
		}
	}
	$res .= "\n"; # -----

	# [OPTIONS]
	$res .= "[OPTIONS]"."\n";
	$res .= "## Format:\n";
	$res .= "# TESTSUITE testsuite [VER >=version] [LSB <=version | Moblin ver]"."\n";
	$res .= "# (ID|GLOBAL): option ID"."\n";
	$res .= "# Q: option title"."\n";
	$res .= "# PRE: perl regular expression (for dialogue with the testsuite)"."\n";
	$res .= "# TYPE: enum (val1, val2) OR bool (val_true, val_false) OR string"."\n";
	$res .= "#       If TYPE is omitted, then TYPE is string by default."."\n";
	$res .= "#       Note, bool type requires two values to be specified."."\n";
	$res .= "# DEFAULT: default value"."\n";
	$res .= "# [TESTSUITE END]"."\n";
	foreach my $testsuite ( sort keys %{$manifest->{OPTIONS}} ) {
		my $options;
		foreach ( @{$manifest->{OPTIONS}{$testsuite}} ) {
			my $test_ver_spec = ( $_->{TEST_VER} ? "VER ".$_->{TEST_VER} : "" );
			my $standard_spec = ( $_->{STANDARD} or "" );
			push @{$options->{$test_ver_spec}{$standard_spec}}, $_;
		}
		
		foreach my $test_ver_spec ( sort keys %$options ) {
		foreach my $standard_spec ( sort keys %{$options->{$test_ver_spec}} ) {
			$res .= "TESTSUITE\t".$testsuite."\t".$test_ver_spec."\t".$standard_spec."\n";
			$res .= "\n";
			
			foreach my $opt ( @{$options->{$test_ver_spec}{$standard_spec}} ) {
				if ( defined $opt->{GLOBAL} && $opt->{GLOBAL} ) {
					$res .= "GLOBAL: ".$opt->{ID}."\n";
				} else {
					$res .= "ID: ".$opt->{ID}."\n";
				}
				$res .= "Q: ".$opt->{Q}."\n" if defined $opt->{Q};
				$res .= "PRE: ".$opt->{PRE}."\n" if defined $opt->{PRE};
				if ( defined $opt->{TYPE} ) {
					$res .= "TYPE: ".$opt->{TYPE};
					if ( defined $opt->{VARIANTS} ) {
					# Print joined option=displayname pairs in parentheses
						my @tmp;
						foreach ( @{$opt->{VARIANTS}} ) {
							my $tmp = $_;
							if ( defined $opt->{DISPLAY} && defined $opt->{DISPLAY}{$_}
								&& $opt->{DISPLAY}{$_} ne $_ ) 
							{
								$tmp .= "=".$opt->{DISPLAY}{$_};
							}
							push @tmp, $tmp;
						}
						$res .= " {".join("|", @tmp)."}";
					}
					$res .= "\n";
				}
				$res .= "DEFAULT: ".$opt->{DEFAULT}."\n" if defined $opt->{DEFAULT};
				$res .= "\n";
			}
			
			$res .= "TESTSUITE END"."\n";
			$res .= "\n"; # -----
		}}
	}

	# [MANUAL]
	$res .= "[MANUAL]"."\n";
	$res .= "## Format:\n";
	$res .= "# TESTSUITE testsuite [LSB >=version | Moblin 2.0]"."\n";
	$res .= "# STEP instructions for tester (in html)"."\n";
	$res .= "# [TESTSUITE END]"."\n";
	foreach my $testsuite ( sort keys %{$manifest->{MANUAL}} ) {
		my $manual;
		foreach ( @{$manifest->{MANUAL}{$testsuite}} ) {
			my $standard_spec = ( $_->{STANDARD} or "" );
			push @{$manual->{$standard_spec}}, $_;
		}
		
		foreach my $standard_spec ( sort keys %$manual ) {
			$res .= "TESTSUITE\t".$testsuite."\t".$standard_spec."\n";
			$res .= "\n";
			
			foreach my $step ( @{$manual->{$standard_spec}} ) {
				$res .= "STEP ".$step->{STEP}."\n";
				$res .= "\n";
			}
			
			$res .= "TESTSUITE END"."\n";
			$res .= "\n"; # -----
		}
	}

	# [PROFILES]
	$res .= "[PROFILES]"."\n";
	$res .= "## Display names for profiles\n";
	$res .= "# Format: PROFILE Standard ver <profile_name> = <profile display name (html)>"."\n";
	# ->{STD_PROFILES}{$standard}{$profile_id} = "Display name (html)"
	foreach my $standard   ( sort keys %{$manifest->{STD_PROFILES}} ) {
	foreach my $profile_name ( sort keys %{$manifest->{STD_PROFILES}{$standard}} ) {
		my $display_name = $manifest->{STD_PROFILES}{$standard}{$profile_name};
		(defined $display_name) or return error "Display name isn't defined for profile '$profile_name' for '$standard'.";
		$res .= "PROFILE ".$standard." ".$profile_name." = ".$display_name;
		$res .= "\n";
	}}
	$res .= "\n"; # -----
	
	$res .= "## Test suite to profile references\n";
	$res .= "# Format: testsuite [Standard ver]: <profiles>"."\n";
	#     ->{PROFILES}{$standard}{$testsuite}{$profile} = 1
	foreach my $standard ( sort keys %{$manifest->{PROFILES}} ) {
	foreach my $testsuite ( sort keys %{$manifest->{PROFILES}{$standard}} ) {
		$res .= $testsuite."\t".$standard.": ";
		$res .= join "|", sort keys %{$manifest->{PROFILES}{$standard}{$testsuite}};
		$res .= "\n";
	}}
	$res .= "\n"; # -----
	
	# [MODULES]
	# ->{MODULES}{$testsuite}[]->{MAIN|FILE|FROM}
	$res .= "[MODULES]"."\n";
	$res .= "# TESTSUITE xts5 [VER >=3.0] LSB|MOBLIN"."\n";
	$res .= "#  module Xts5_test.pm	[from /var/opt/lsb/.../Xts5_test.pm]"."\n";
	$res .= "#  requires X11_common.pm	from http://.../X11_common.pm"."\n";
	$res .= "\n";
	foreach my $testsuite ( sort keys %{$manifest->{MODULES}} ) {
		my $modules;
		foreach ( @{$manifest->{MODULES}{$testsuite}} ) {
			my $test_ver_spec = ( $_->{TEST_VER} ? "VER ".$_->{TEST_VER} : "" );
			my $standard_spec = ( $_->{STANDARD} or "" );
			push @{$modules->{$test_ver_spec}{$standard_spec}}, $_;
		}
		
		foreach my $test_ver_spec ( sort keys %$modules ) {
		foreach my $standard_spec ( sort keys %{$modules->{$test_ver_spec}} ) {
			$res .= "TESTSUITE\t".$testsuite."\t".$test_ver_spec."\t".$standard_spec."\n";
			
			foreach my $obj ( @{$modules->{$test_ver_spec}{$standard_spec}} ) {
				$res .= " ".($obj->{MAIN}?'module':'requires')." ".$obj->{FILE};
				$res .= " from ".$obj->{FROM} if defined $obj->{FROM};
				$res .= "\n";
			}
		}}
		$res .= "\n";
	}
	
	# [COMPOSITION]
	$res .= "[COMPOSITION]"."\n";
	$res .= "## Format:\n";
	$res .= "# TESTSUITE  test\tversion\tStandard ver [arch] status"."\n";
	$res .= "#  package-name\tpackage-ver\t[location] (comma-separated and/or on a new line)"."\n";
	$res .= "\n";
	#  ->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}{$test_ver}[]->{PACKAGE|PACKAGE_VER|LOCATION} = ".."
	foreach my $standard ( sort keys %{$manifest->{COMPOSITION}} ) {
	foreach my $testsuite   ( sort keys %{$manifest->{COMPOSITION}{$standard}} ) {
		my $arch_ref = $manifest->{COMPOSITION}{$standard}{$testsuite};
		
		# Pack architectures together
		my $aggreg_hash = {};

		foreach my $arch ( sort keys %$arch_ref ) {
		foreach my $status ( keys %{$arch_ref->{$arch}} ) {
		foreach my $test_ver ( keys %{$arch_ref->{$arch}{$status}} ) {
			my $reqs = $arch_ref->{$arch}{$status}{$test_ver} or next;
			my $key = "";
			foreach my $package ( @$reqs ) {
				$key .= $package->{PACKAGE};
				$key .= "-".$package->{PACKAGE_VER} if defined $package->{PACKAGE_VER};
				$key .= "@".$package->{LOCATION} if defined defined $package->{LOCATION};
				$key .= "\t";
			}
			
			foreach my $arch_d ( @all_archs ) { # expand arch *
				next if $arch ne '*' && $arch_d ne $arch;
				
				$aggreg_hash->{$key}{$test_ver}{$status}{$arch_d} = 1;
			}
		}}}
		
		foreach my $key ( sort keys %$aggreg_hash ) {
		foreach my $test_ver ( sort keys %{$aggreg_hash->{$key}} ) {
		foreach my $status ( sort keys %{$aggreg_hash->{$key}{$test_ver}} ) {
			my @archs = keys %{$aggreg_hash->{$key}{$test_ver}{$status}};
			
			my $arch_str = "";
			if ( scalar @archs == scalar @all_archs ) {
				$arch_str = "";
			} else {
				$arch_str = (join ",", @archs)."\t";
			}
		
			$res .= "TESTSUITE ".$testsuite."\t".$test_ver."\t\t".$standard;
			$res .= "\t".$arch_str;
			$res .= $status."\n";
			
			my $any_arch = $archs[0]; # Should have equal reqs
			my $reqs = $arch_ref->{$any_arch}{$status}{$test_ver};
		
			foreach my $req ( @$reqs ) {
				$res .= "\t".$req->{PACKAGE};
				$res .= "\t".$req->{PACKAGE_VER} if defined $req->{PACKAGE_VER} && $req->{PACKAGE_VER} ne "";
				$res .= "\t".$req->{LOCATION} if defined $req->{LOCATION} && $req->{LOCATION} ne "";
				$res .= "\n";
			}
			$res .= "\n";
		}}}
	}}
	$res .= "\n"; # -----

	# [PACKAGES]
	$res .= "[PACKAGES]"."\n";
	if ( defined $manifest->{CONFLICTS} ) {
		my $conflicts = { %{$manifest->{CONFLICTS}} };
		
		while ( %$conflicts ) {
			my $seed = (sort keys %$conflicts)[0];
			
			my @cluster = ($seed);
			
			INNER:
			while ( 1 ) {
				foreach ( @cluster ) {
					if ( defined $conflicts->{$_} ) {
						push @cluster, grep {!in_array($_, @cluster) } keys %{$conflicts->{$_}};
						delete $conflicts->{$_};
						next INNER;
					}
				}
				last;
			}
			
			$res .= "CONFLICTS ".join(", ", @cluster)."\n";
		}
	}
	$res .= "\n"; # -----

	# [SERVERS]
	$res .= "[SERVERS]"."\n";
	$res .= "## Format:\n";
	$res .= "# DIR dir_in_cache"."\n";
	$res .= "#   url_corresponding_to_cache_dir"."\n";
	foreach my $dir ( keys %{$manifest->{SERVERS}} ) {
		$res .= "DIR $dir"."\n";
		foreach my $link ( keys %{$manifest->{SERVERS}{$dir}} ) {
			$res .= "\t".$link."\n";
		}
		$res .= "\n";
	}
	$res .= "\n"; # -----
		
	# [FILES]
	$res .= "[FILES]"."\n";

	foreach my $name ( sort keys %{$manifest->{LOCATIONS}} ) {
		$res .= "ALIAS ".$name." = ".$manifest->{LOCATIONS}{$name}."\n";
	}
	$res .= "\n";
	
	$res .= "## Format:\n";
	$res .= "# filename | location | size [timestamp]"
		." [ arch [ package kind | package name | package version ]]"."\n";
	foreach my $key ( sort keys %{$manifest->{FILES}} ) {
		my $file = $manifest->{FILES}{$key};
		
		$res .= print_file_line($file)."\n";
	}
	$res .= "\n"; # -----
	
	return $res;
}
#----------------------------------------------------------------------

sub locally_available {
	my ($manifest) = @_;
	
	foreach my $filekey ( keys %{$manifest->{FILES}} ) {
		my $file = $manifest->{FILES}{$filekey};
		my $file_in_cache = cache_dest($file->{SRC})."/".$file->{FILENAME};
		
		if ( -f $file_in_cache ) {
			$file->{USED} = 1;
			$file->{REMOVED} = 1; # TODO: Describe it
		}
	}
	
	remove_unavailable($manifest);
	
	# Change cert status to certold
	foreach my $standard ( keys %{$manifest->{COMPOSITION}} ) {
	foreach my $testsuite ( keys %{$manifest->{COMPOSITION}{$standard}} ) {
	foreach my $arch ( keys %{$manifest->{COMPOSITION}{$standard}{$testsuite}} ) {
		my $statuses = $manifest->{COMPOSITION}{$standard}{$testsuite}{$arch};
		if ( defined $statuses->{'cert'} ) {
			$statuses->{'certold'} = $statuses->{'cert'};
			delete $statuses->{'cert'};
		}
	}}}
	
	return $manifest;
}

#-----------------------------------------------------------------------

# Syncronize Manifest with the server
#
# $manif_file - Manifest file name with path. Manifest should contain URL which is used for updating.
# If there is no such file, or Manifest doesn't contain the URL, then the $url_res is used.
sub update_manifest {
	my ($manif_file, $url_res) = @_;
	
	my $manifest;
	if ( -f $manif_file ) {
		$manifest = manifest_parse( $manif_file );
		if ( !is_ok($manifest) ) {
			# The error returned by manifest_parse() should contain
			# the part of the Manifest it has managed to parse.
			$manifest = $Error::Last->{-data};
		}
	}
	
	# Take the Manifest URL from the existing Manifest header
	my $url;
	if ( $manifest && $manifest->{MANIFEST}{URL} ) {
		$url = $manifest->{MANIFEST}{URL};
	}
	if ( !$url ) {
		$url = $url_res;
	}
	if ( !$url ) {
		return error "Manifest doesn't contain URL for updating. Please try to specify it manually.";
	}
	
	# Download new Manifest file
	my $manif_new_file = $manif_file.".new";
	if ( -f $manif_new_file ) { # Delete this file if it is on the way
		cmd( "rm -f ".shq($manif_new_file) );
		( !-f $manif_new_file ) # Check if deleted
			or return error "Failed to remove '$manif_new_file'", error $last_cmd_output;
	}
	
	is_ok download_file($url, $manif_new_file)
		or return error "Failed to download a new Manifest file", $Error::Last;
	
	# check if the file was really downloaded
	( -f $manif_new_file && !-z $manif_new_file) # File exists and not empty
		or return error "Failed to download a new Manifest file";
	
	if ( $manifest ) {
		# Backup the Manifest file
		cmd( "cp -f ".shq($manif_file)." ".shq($manif_file.".old") ) == 0
			or return error "Failed to backup the Manifest file";
		
		# Merge new manifest with the local one
		my $manifest2 = manifest_parse( $manif_new_file );
		is_ok($manifest2) or return error "Failed to parse the manifest file '$manif_new_file'", $Error::Last;
		
		# Filter locally available tests
		locally_available($manifest);
		
		is_ok merge_manifests($manifest, $manifest2)
			or return error "Conflict during merge, fail.", $Error::Last, -ret => 2;
		
		$manifest->{MANIFEST}{TIMESTAMP} = time();
		
		is_ok check_manifest($manifest)
			or return error("Manifest check failed", $Error::Last->{-arr}[0], -ret => 2);
		
		# Write a new Manifest
		is_ok write_string_as_file( "!".$manif_file, print_manifest($manifest) )
			or return error "Failed to write the Manifest file", $Error::Last, -ret => 2;
		
		inform "Manifest has been written to: '$manif_file'\n";
	}
	else {
		# Check the downloaded Manifest
		my $manifest = manifest_parse( $manif_new_file );
		is_ok($manifest) or return error "Failed to parse manifest '$manif_new_file'", $Error::Last, -ret => 2;
		
		$manifest->{MANIFEST}{TIMESTAMP} = time();
		
		is_ok check_manifest( $manifest )
			or return error("Manifest check failed", $Error::Last->{-arr}[0], -ret => 2);
		
		# Write a new Manifest
		is_ok write_string_as_file( $manif_file, print_manifest($manifest) )
			or return error "Failed to write the Manifest file", $Error::Last, -ret => 2;
		
		inform "Manifest has been written to: '$manif_file'\n";
	}
	
	return 1;
}
#-----------------------------------------------------------------------

sub rpm_deps {
	my ($rpm_file) = @_;
	
	my @requires = ();
	
	my $qR = capture("rpm -qR -p ".shq($rpm_file)." 2>/dev/null");
	chomp $qR;
	foreach my $line ( split /\n/, $qR ) {
		$line =~ s/ .*//; # remove version
		$line =~ s/\(.*//; # remove stuff in parentheses

		push @requires, $line;
	}
	
	return @requires;
}

sub check_deps {
	my ($directory, $manifest) = @_;
	
	my $errors = 0;
	
	my $rpm_info_cache = {};
	
	#  ->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}{$test_ver}[]->{PACKAGE|PACKAGE_VER|LOCATION} = ".."
	my $composition = $manifest->{COMPOSITION};
	foreach my $standard  ( keys %$composition ) {
	foreach my $testsuite ( keys %{$composition->{$standard}} ) {
	foreach my $arch      ( keys %{$composition->{$standard}{$testsuite}} ) {
	foreach my $status    ( keys %{$composition->{$standard}{$testsuite}{$arch}} ) {
	foreach my $test_ver  ( keys %{$composition->{$standard}{$testsuite}{$arch}{$status}} ) {
		inform "> '$standard' - '$testsuite', '$arch', '$status', ver '$test_ver'";
		my $reqs = $composition->{$standard}{$testsuite}{$arch}{$status}{$test_ver};
		my %tmp = ();
		foreach my $req ( @$reqs ) {
			my $package_name = $req->{PACKAGE};
			my $package_ver  = ( $req->{PACKAGE_VER} or "*" );
			my $location     = ( $req->{LOCATION} or "" );
			
			if ( $package_ver eq "*" ) {
				( $location ne "" ) or warning "Location should be defined for template";
			}
			
			my $package_manager = "rpm";

			my $found_files = Manifest::select_file( $manifest, $arch, $package_manager,
						$package_name, $package_ver, $location);
			
			if ( is_err($found_files) || !$found_files || !@$found_files ) {
				$errors = 1;
				warning "check: Can't find file '$package_name' v. '$package_ver'"
					." ($package_manager), '$standard', arch: '$arch',"
					." required by '$testsuite'"
					.($location ? " in '$location'": ""), $Error::Last;
				next;
			}
			if ( @$found_files > 1 ) {
				$errors = 1;
				warning "Ambigous file choise: ".$found_files->[0]->{SRC}
					.": ".join ", ", map {$_->{FILENAME}} @$found_files;
			}
			
			if ( $found_files->[0]->{KIND} eq 'file' ) {
				$tmp{$package_name} = 2;
				next;
			}
			( $found_files->[0]->{KIND} eq 'rpm' ) or die;
			
			my $fullname = $found_files->[0]->{SRC}."/".$found_files->[0]->{FILENAME};
			$fullname =~ s{^/[^/]*}{}; # remove root dir (domain name)
			$fullname = $directory.$fullname;
			
			if ( !-f $fullname ) {
				$errors = 1;
				warning "Missing file '$fullname'";
				$tmp{$package_name} = 0;
				next;
			}
			
			my @deps;
			if ( defined $rpm_info_cache->{$fullname} ) { # look in the cache first
				@deps = @{$rpm_info_cache->{$fullname}};
			} else {
				@deps = rpm_deps($fullname); # call `rpm` to get dependencies of the package.
				$rpm_info_cache->{$fullname} ||= [@deps]; # cache results
			}
			
			foreach my $dep ( @deps ) {
				next if $dep =~ /\/bin\/sh/;
				next if $dep =~ /^rpmlib(\s|$)/;
				next if $dep =~ /^lsb(\s|$)/;
				next if $dep =~ /^lsb-(core|graphics)-/;
				next if $dep =~ /(^|\/)perl(\s|$)/;
				next if $dep =~ /^lib.*\.so/;
				
				if ( defined $tmp{$dep} ) {
					$tmp{$dep} = 2; # 2 means 'used';
				} else {
					$errors = 1;
					warning "'$testsuite', $standard, '$status', '$arch', ver '$test_ver':"
						." '$fullname' ($package_name) \e[0;31mneeds '$dep'.\e[0m";
					inform "Req: ".join ", ", map {$_->{PACKAGE}} @$reqs;
				}
			}
			$tmp{$package_name} = 1 if !$tmp{$package_name};
		} # end of  foreach my $req ( @$reqs )
		
		# Check unneeded reqs
		next if !$tmp{$reqs->[-1]->{PACKAGE}};
		foreach my $req ( @$reqs[0..($#$reqs-1)] ) {
			my $package_name = $req->{PACKAGE};
			#( $tmp{$package_name} ) or die $package_name;
			if ( $tmp{$package_name} != 2 && $package_name ne 'lsb-setup' ) {
				$errors = 1;
				warning "'$testsuite', $standard, '$status', '$arch', ver '$test_ver':"
						." \e[0;33mdoesn't need '$package_name'.\e[0m";
			}
		}
	}}}}} # end of  foreach {$testsuite}{$standard}{$status}{$arch}{$test_ver}
	
	return !$errors;
}

#-----------------------------------------------------------------------

sub force_time {
	my $manifest = load_manifest();
	is_ok($manifest) or return error "Failed to load Manifest.", $Error::Last;
	
	foreach my $key ( sort keys %{$manifest->{FILES}} ) {
		my $file = $manifest->{FILES}{$key};
		my $timestamp = $file->{TIMESTAMP} or next;
		my $file_in_cache = cache_dest($file->{SRC})."/".$file->{FILENAME};
		if ( -f $file_in_cache ) {
			inform "> $file->{SRC}  $timestamp";
			utime $timestamp, $timestamp, $file_in_cache;
		}
		else {
			#substr($file_in_cache, 0, length($file_in_cache)-100) = "...";
			$file_in_cache =~ s{.*/pub/lsb/}{};
			inform "X $file_in_cache  $timestamp";
		}
	}
	
	return 1;
}
#-----------------------------------------------------------------------

sub find_files {
	my ($dir) = @_;
	
	my @result = ();
	
	# Open the directory
	opendir DIR, $dir or fail "Can't open directory '$dir': $!";
	
	# Read the directory for files and dirs
	my @files = readdir(DIR);
	closedir DIR;
	
	foreach my $file ( @files ) {
		next if $file =~ /^\./; # skip hidden files, '.' and '..'
		
		my $fullname = "$dir/$file";
		if ( -f $fullname ) {   # a file
			$fullname =~ s{//+}{/};
			push @result, $fullname;
		}
		elsif ( -d $fullname ) { # a directory
			# Do recursive search
			push @result, find_files($fullname);
		}
	}
	
	return @result;
}

sub old_files {
	my ($remove) = @_;
	
	my $manifest = load_manifest();
	is_ok($manifest) or return error "Failed to load Manifest", $Error::Last;
	
	my $files = {};
	
	local $_;
	$files->{$_} = 0 foreach ( find_files(cache_dest()) );
	
	for my $key ( keys %{$manifest->{FILES}} ) {
		my $file = $manifest->{FILES}{$key};
		my $file_in_cache = cache_dest($file->{SRC})."/".$file->{FILENAME};
		$file_in_cache =~ s{//+}{/};
		#inform "? $file_in_cache";
		if ( defined $files->{$file_in_cache} ) {
			$files->{$file_in_cache} = 1;
		}
	}
	foreach my $file_in_cache ( sort keys %$files ) {
		if ( $files->{$file_in_cache} == 0 ) {
			inform "- $file_in_cache";
			if ( $remove ) {
				cmd "rm -f ".shq($file_in_cache);
			}
		}
	}
}
#-----------------------------------------------------------------------

sub setup_download {
	$Download::enabled = 1;

	if ( $Download::tool eq 'curl' ) {
		if( !does_shell_know("curl") ) {
			if( !does_shell_know("wget") ) { # try wget instead of curl
				fail ( "Couldn't find neither 'curl' nor 'wget'." );
			} else {
				inform "'curl' is unavailable, using 'wget' instead.";
				$Download::tool = 'wget';
			}
		}
	}
	elsif ( $Download::tool eq 'wget' ) {
		if( !does_shell_know("wget") ) {
			if( !does_shell_know("curl") ) { # try curl instead of wget
				fail ( "Couldn't find neither 'curl' nor 'wget'." );
			} else {
				inform "'wget' is unavailable, using 'curl'.";
				$Download::tool = 'curl';
			}
		}
	}
}

#-----------------------------------------------------------------------

sub print_help {
	
	# These options are obsolete
	#--dir <directory>   Specify the directory with packages.
	#--ftp <host/dir>    Specify the FTP source of packages.
	#--refresh           Drop all cached FTP lists.
	
	print <<HEREDOC;
Usage: $0 OPTIONS
Manipulate Manifest files.

Options:
  
  --update [file [url]]   Update the Manifest file from the server.
  -o <filename>           Specify the output filename.
  --loc <Manifest>        Extract locally available tests from the Manifest file.
  --merge <Manifest1>     <Manifest2>   Merge two manifests into one.
  --check <Manifest>      Checks the Manifest consistency.
  --check-deps            Checks package dependencies. The package dir should be
                          specified with the --dir option.
  --force-time            Set timestamps for all files in the package cache
                          as specified in the Manifest.
  --old-files             List all files that are no longer in the Manifest.
  --old-files-rm          Remove all files that are no longer in the Manifest.

Another usage: $0 -o Manifest TEMPLATES
Generate Manifest from the template files.

HEREDOC
}

#-----------------------------------------------------------------------

## BEGIN ##

$Ftp::list_refresh = 0;

# Read args
if ( @ARGV < 1 ) {
	print_help();
	exit;
}
my @args = @ARGV;

my @manif_files = ();

while ( @args ) {
	my $arg = shift @args;
	
	if ( $arg eq "--help" ) {
		print_help();
		exit 0;
	}
	elsif ( $arg eq "--refresh" ) {
		$Ftp::list_refresh = 1;
	}
	elsif ( $arg eq "-o" ) {
		$output_filename = shift @args;
	}
	elsif ( $arg eq '--dir' ) {
		$directory = shift @args or fail "Directory name expected after $arg";
	}
	elsif ( $arg eq '--ftp' ) {
		$ftp = 1;
	}
	elsif ( $arg eq '--loc' ) {
		my $manif_file = shift @args or fail "Manifest file is expected after $arg";
		( defined $output_filename ) or fail "Please specify the manifest file name using the -o key before $arg";
		
		my $manifest = manifest_parse( $manif_file );
		is_ok($manifest) or fail "Failed to parse manifest '$manif_file'", $Error::Last;
		
		locally_available($manifest);
		
		is_ok check_manifest($manifest)
			or fail "Manifest check failed", ($Error::Last->{-arr} ? $Error::Last->{-arr}[0] : $Error::Last);
		
		is_ok write_string_as_file($output_filename, print_manifest($manifest))
			or warning "Failed to write $output_filename", $Error::Last;
		
		print "Manifest was written to: $output_filename\n";
		
		exit 0;
	}
	elsif ( $arg eq '--merge' ) {
		my $manif_file1 = shift @args or fail "Two Manifest files are expected after $arg";
		my $manif_file2 = shift @args or fail "Two Manifest files are expected after $arg";
		( defined $output_filename ) or fail "Please specify the manifest file name using the -o key before $arg";
		
		my $manifest = manifest_parse( $manif_file1 );
		is_ok($manifest) or fail "Failed to parse the manifest file '$manif_file1'.", $Error::Last;
		
		my $manifest2 = manifest_parse( $manif_file2 );
		is_ok($manifest2) or fail "Failed to parse the manifest file '$manif_file2'.", $Error::Last;
		
		is_ok merge_manifests($manifest, $manifest2)
			or fail "Conflict during merge, fail.", $Error::Last;
		
		is_ok check_manifest( $manifest ) or complain $Error::Last;
		
		is_ok write_string_as_file($output_filename, print_manifest($manifest))
			or warning "Failed to write $output_filename", $Error::Last;
		print "Manifest was written to: $output_filename\n";
		exit 0;
	}
	elsif ( $arg eq '--check' ) {
		my $manif_file = shift @args;
		if ( !defined $manif_file ) { $manif_file = manifest_dir."/Manifest" };
		
		my $manifest = manifest_parse( $manif_file );
		is_ok($manifest) or fail "Failed to parse the manifest file '$manif_file'.", $Error::Last;
		
		is_ok check_manifest($manifest) or fail $Error::Last;
		
		exit 0;
	}
	elsif ( $arg eq '--check-deps' ) {
		my $manif_file = shift @args or fail "Manifest file is expected after $arg";
		
		( $directory && -d $directory ) or fail "Please specify the directory using the --dir key before $arg";
		
		my $manifest = manifest_parse( $manif_file );
		is_ok $manifest or fail "Failed to parse the manifest file '$manif_file'.", $Error::Last;
		
		check_deps($directory, $manifest)
			or fail "Dependency check found errors.";
		
		exit 0;
	}
	elsif ( $arg eq '--update' ) { # --update [file [url]]
		my $manif_file = shift @args; # Optional parameter
		my $url_res = shift @args; # Optional parameter

		my @files = ();
		if ( $manif_file ) {
			push @files, $manif_file;
		} else {
			$url_res = $DEFAULT_MANIFEST_URL;
			push @files, manifest_dir()."/Manifest";
			push @files, glob(manifest_dir()."/*.mnf");
		}

		setup_download();

		my $error = 0;
		for my $manif_file ( @files ) {
			is_ok update_manifest($manif_file, $url_res)
			or do {
				complain "Failed to update '$manif_file'", $Error::Last;
				$error = $Error::Last->{-ret} || 1;
			};
		}
		exit($error);
	}
	elsif ( $arg eq '--force-time' ) {
		is_ok force_time() or fail "Failed to apply timestamps";
		exit 0;
	}
	elsif ( $arg eq '--old-files' ) {
		is_ok old_files() or fail $Error::Last;
		exit 0;
	}
	elsif ( $arg eq '--old-files-rm' ) {
		is_ok old_files(-remove) or fail $Error::Last;
		exit 0;
	}
	else {
		if ( $arg =~ /^-/ ) {
			complain "Unknown option: '$arg'";
			print_help();
			exit 1;
		}

		my @files = glob $arg
			or fail "No such file(s): '$arg'";
		
		push @manif_files, @files;
	}
}

#( defined $output_filename ) or fail "Please specify the manifest file name using the -o key";

# MAIN
my $manifest = main( @manif_files );
is_ok $manifest or fail "Failed to generate manifest", $Error::Last;
#

my $manif_text = print_manifest($manifest);

if ( defined $output_filename ) {
	is_ok write_string_as_file($output_filename, $manif_text)
		or warning "Failed to write $output_filename", $Error::Last;
	
	print "Manifest was written to: $output_filename\n";
	
	# Re-parse the result %)
	my $manifest2 = manifest_parse( $output_filename );
	is_ok($manifest2) or fail "Failed to parse '$output_filename'", $Error::Last;
	
	my $manif_text2 = print_manifest($manifest2);
	
	if ( $manif_text ne $manif_text2 ) {
		is_ok write_string_as_file( $output_filename.".reparsed", $manif_text2 )
			or warning "Failed to write ${output_filename}.reparsed", $Error::Last;
		
		print STDERR "--- DIFF: ---\n";
		
		system ( "diff '$output_filename' '$output_filename.reparsed'" );
		
		fail "Re-parsed manifest differs from the original one!";
	}
	print "Manifest was successfully generated: $output_filename\n";
} else {
	print $manif_text;
}

exit 0;
